import os
from collections.abc import Callable
from datetime import timedelta

import pytest
from dagster import AssetKey, DagsterInstance
from dagster._time import get_current_datetime
from dagster_airlift.core.airflow_instance import AirflowInstance


@pytest.fixture(name="dagster_home")
def dagster_home_fixture(local_env: None) -> str:
    return os.environ["DAGSTER_HOME"]


@pytest.fixture(name="module_and_instance")
def module_and_instance_fixture(request) -> str:
    return request.param


@pytest.fixture(name="dagster_dev_cmd")
def dagster_dev_cmd_fixture(module_and_instance: tuple) -> list[str]:
    module, _instance = module_and_instance
    return ["dagster", "dev", "-m", module, "-p", "3333"]


def peer_instance() -> AirflowInstance:
    from perf_harness.dagster_defs.peer import airflow_instance as af_instance_peer

    return af_instance_peer


def observe_instance() -> AirflowInstance:
    from perf_harness.dagster_defs.observe import airflow_instance as af_instance_observe

    return af_instance_observe


def migrate_instance() -> AirflowInstance:
    from perf_harness.dagster_defs.migrate import airflow_instance as af_instance_migrate

    return af_instance_migrate


@pytest.mark.parametrize(
    "module_and_instance",
    [
        ("perf_harness.dagster_defs.peer", peer_instance),
        ("perf_harness.dagster_defs.observe", observe_instance),
        ("perf_harness.dagster_defs.migrate", migrate_instance),
    ],
    ids=["peer", "observe", "migrate"],
    indirect=True,
)
def test_dagster_materializes(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
    module_and_instance: tuple[str, Callable[[], AirflowInstance]],
) -> None:
    """Test that assets can load properly, and that materializations register."""
    module, instance_fn = module_and_instance
    af_instance = instance_fn()
    run_id = af_instance.trigger_dag("dag_0")
    af_instance.wait_for_run_completion(dag_id="dag_0", run_id=run_id, timeout=60)
    dagster_instance = DagsterInstance.get()
    start_time = get_current_datetime()
    while get_current_datetime() - start_time < timedelta(seconds=60):
        asset_materialization = dagster_instance.get_latest_materialization_event(
            asset_key=AssetKey(["my_airflow_instance", "dag", "dag_0"])
        )
        if asset_materialization:
            break

    assert asset_materialization, "Could not find asset materialization for asset dag key"  # pyright: ignore[reportPossiblyUnboundVariable]

    if module.endswith("observe") or module.endswith("migrate"):
        asset_materialization = dagster_instance.get_latest_materialization_event(
            asset_key=AssetKey(["task_0_0_asset_0"])
        )
        assert asset_materialization



import subprocess
import tempfile
from pathlib import Path

import pytest


@pytest.fixture
def temp_venv():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_venv = Path(temp_dir) / "venv"
        try:
            # Check if uv is installed
            subprocess.run(["uv", "--version"], check=True, capture_output=True)
        except (subprocess.CalledProcessError, FileNotFoundError):
            raise Exception("uv is not installed or not in PATH")

        # Create a new virtual environment using uv
        result = subprocess.run(
            ["uv", "venv", temp_dir_venv, "--seed"], check=True, capture_output=True, text=True
        )
        assert result.returncode == 0, f"Failed to create virtual environment: {result.stderr}"

        # Verify the environment was created by checking if python executable exists
        python_executable = temp_dir_venv / "bin" / "python"
        assert python_executable.exists(), f"Python executable not found at {python_executable}"
        # Install uv
        uv_install_output = subprocess.run(
            [python_executable, "-m", "pip", "install", "uv"],
            check=True,
            capture_output=True,
            text=True,
        )
        assert uv_install_output.returncode == 0, (
            f"Failed to install uv: {uv_install_output.stderr}"
        )
        yield temp_dir_venv


def test_in_airflow_package_implicit_requirements(temp_venv: Path):
    """The dagster-airlift[in-airflow] package has an implicit dependency on Airflow, and should be able to work without dagster installed.
    This test news up a fresh virtualenv, installs the dagster-airlift[in-airflow] package, and ensures the following:
    - import dagster fails
    - import dagster_airlift.in_airflow fails with an error message indicating that Airflow is not installed
    Then, we install Airflow properly and run a separate import script to ensure that the package works as expected, and does not require dagster to be installed.
    """
    path_to_package = Path(__file__).parent.parent.parent.parent
    # install package using uv pip install
    python_executable = temp_venv / "bin" / "python"
    uv_entrypoint = [python_executable, "-m", "uv"]
    install_result = subprocess.run(
        [*uv_entrypoint, "pip", "install", "-e", ".[in-airflow]"],
        check=True,
        capture_output=True,
        text=True,
        cwd=path_to_package,
    )
    assert install_result.returncode == 0, (
        f"Failed to install package in virtual environment: {install_result.stderr}"
    )

    uv_pip_freeze_output = subprocess.run(
        [*uv_entrypoint, "pip", "freeze"], check=True, capture_output=True, text=True
    )
    assert install_result.returncode == 0, (
        f"Failed to get pip freeze output: {uv_pip_freeze_output.stderr}"
    )
    assert "dagster-airlift" in uv_pip_freeze_output.stdout, (
        "dagster-airlift not found in pip freeze output"
    )

    # Run the import script
    import_script_path = Path(__file__).parent / "import_script.py"
    script_result = subprocess.run(
        [python_executable, import_script_path], check=True, capture_output=True, text=True
    )
    assert script_result.returncode == 0, f"Script execution failed: {script_result.stderr}"

    # Install Airflow
    airflow_install_result = subprocess.run(
        [
            *uv_entrypoint,
            "pip",
            "install",
            "apache-airflow",
            # https://github.com/apache/airflow/discussions/57769#discussioncomment-14861217
            "fastapi<0.118",
            # https://github.com/apache/airflow/issues/57419
            "structlog<25.5.0",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    assert airflow_install_result.returncode == 0, (
        f"Failed to install Airflow: {airflow_install_result.stderr}"
    )

    import_airflow_script_path = Path(__file__).parent / "import_script_with_airflow.py"
    airflow_script_result = subprocess.run(
        [python_executable, import_airflow_script_path], check=True, capture_output=False, text=True
    )
    assert airflow_script_result.returncode == 0, (
        f"Script execution failed: {airflow_script_result.stderr}"
    )



from pathlib import Path
from typing import Any

import pytest
import requests


def assert_link_exists(link_name: str, link_url: Any):
    assert isinstance(link_url, str)
    assert requests.get(link_url).status_code == 200, f"{link_name} is broken"


@pytest.fixture(name="dags_dir")
def default_dags_dir():
    return Path(__file__).parent / "dags"



from dagster_airlift.in_airflow.task_proxy_operator import matched_dag_id_task_id


def test_parse_asset_node() -> None:
    asset_node = {
        "id": 'dbt_example.dagster_defs.migrate.__repository__.["lakehouse", "iris"]',
        "assetKey": {"path": ["lakehouse", "iris"]},
        "metadataEntries": [
            {
                "label": "dagster-airlift/task-mapping",
                "jsonString": '[{"dag_id": "rebuild_iris_models", "task_id": "load_iris"}]',
                "__typename": "JsonMetadataEntry",
            },
            {
                "label": "Task Info (raw)",
                "jsonString": '{"class_ref": {"class_name": "DefaultProxyToDagsterOperator", "module_path": "dagster_airlift.in_airflow.base_proxy_operator"}, "depends_on_past": false, "downstream_task_ids": ["build_dbt_models"], "end_date": null, "execution_timeout": null, "extra_links": [], "is_mapped": false, "operator_name": "DefaultProxyToDagsterOperator", "owner": "airflow", "params": {}, "pool": "default_pool", "pool_slots": 1.0, "priority_weight": 1.0, "queue": "default", "retries": 0.0, "retry_delay": {"__type": "TimeDelta", "days": 0, "microseconds": 0, "seconds": 300}, "retry_exponential_backoff": false, "start_date": "2024-09-24T18:48:05.148806+00:00", "task_id": "load_iris", "template_fields": [], "trigger_rule": "all_success", "ui_color": "#fff", "ui_fgcolor": "#000", "wait_for_downstream": false, "weight_rule": "downstream"}',
                "__typename": "JsonMetadataEntry",
            },
            {"label": "Dag ID", "text": "rebuild_iris_models", "__typename": "TextMetadataEntry"},
            {"__typename": "UrlMetadataEntry"},
            {
                "label": "Triggered by Task ID",
                "text": "load_iris",
                "__typename": "TextMetadataEntry",
            },
        ],
        "jobs": [
            {
                "id": "50b0b23cdb52e8f68e39bce56dd28e30ef221a53",
                "name": "__ASSET_JOB",
                "repository": {
                    "id": "0abb753321e78ef093132df5e05aa2046ce7bca9::7170e8bbd3d71e16435c1fbdad3185ab83074660",
                    "name": "__repository__",
                    "location": {
                        "id": "dbt_example.dagster_defs.migrate",
                        "name": "dbt_example.dagster_defs.migrate",
                    },
                },
            }
        ],
    }
    assert matched_dag_id_task_id(asset_node, "rebuild_iris_models", "load_iris")
    assert not matched_dag_id_task_id(asset_node, "other_dag", "other_task")



import time
from datetime import timedelta

import pytest
from dagster import AssetKey, DagsterInstance, DagsterRunStatus
from dagster._core.definitions import materialize
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.metadata.metadata_value import JsonMetadataValue
from dagster._core.events.log import EventLogEntry
from dagster._core.storage.dagster_run import RunsFilter
from dagster._time import get_current_datetime, get_current_datetime_midnight
from dagster_airlift.in_airflow.base_asset_operator import DAG_RUN_ID_TAG_KEY
from kitchen_sink.airflow_instance import local_airflow_instance

from kitchen_sink_tests.integration_tests.conftest import (
    makefile_dir,
    poll_for_airflow_run_existence_and_completion,
    poll_for_expected_mats,
    poll_for_materialization,
)


@pytest.fixture(name="dagster_dev_cmd")
def dagster_dev_cmd_fixture() -> list[str]:
    return ["make", "run_dagster_mapped", "-C", str(makefile_dir())]


def test_migrated_dagster_print_materializes(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that assets can load properly, and that materializations register."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()

    expected_mats_per_dag = {
        "print_dag": [AssetKey("print_asset")],
    }

    poll_for_expected_mats(af_instance, expected_mats_per_dag)


RAW_METADATA_KEY = "Run Metadata (raw)"


def dag_id_of_mat(event_log_entry: EventLogEntry) -> bool:
    assert event_log_entry.asset_materialization
    assert isinstance(event_log_entry.asset_materialization.metadata, dict)
    json_metadata_value = event_log_entry.asset_materialization.metadata[RAW_METADATA_KEY]
    assert isinstance(json_metadata_value, JsonMetadataValue)
    assert isinstance(json_metadata_value.data, dict)
    return json_metadata_value.data["dag_id"]


def test_dagster_weekly_daily_materializes(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that asset orchestrated by two dags loads property. Then
    it triggers both dags that target it, and ensure that two materializations
    register.
    """
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()

    dag_id = "weekly_dag"
    asset_one = AssetKey("asset_one")
    dag_run_id = af_instance.trigger_dag(dag_id=dag_id)
    af_instance.wait_for_run_completion(dag_id=dag_id, run_id=dag_run_id, timeout=60)
    dagster_instance = DagsterInstance.get()

    dag_asset_key = AssetKey(["my_airflow_instance", "dag", dag_id])
    assert poll_for_materialization(dagster_instance, dag_asset_key)
    weekly_mat_event = poll_for_materialization(dagster_instance, asset_one)
    assert weekly_mat_event.asset_materialization
    assert weekly_mat_event.asset_materialization.asset_key == asset_one
    assert dag_id_of_mat(weekly_mat_event) == "weekly_dag"

    dag_id = "daily_dag"
    dag_run_id = af_instance.trigger_dag(dag_id=dag_id)
    af_instance.wait_for_run_completion(dag_id=dag_id, run_id=dag_run_id, timeout=60)

    start_time = get_current_datetime()
    final_result = None
    while get_current_datetime() - start_time < timedelta(seconds=30):
        records_result = dagster_instance.fetch_materializations(records_filter=asset_one, limit=10)

        if len(records_result.records) == 2:
            final_result = records_result
            break

        time.sleep(0.1)

    assert final_result, "Did not get two materializations and timed out"

    assert final_result.records[0].event_log_entry
    assert dag_id_of_mat(final_result.records[0].event_log_entry) == "daily_dag"
    assert dag_id_of_mat(final_result.records[1].event_log_entry) == "weekly_dag"


def test_migrated_overridden_dag_materializes(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that assets are properly materialized from an overridden dag."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()

    expected_mats_per_dag = {
        "overridden_dag": [AssetKey("asset_two")],
    }
    poll_for_expected_mats(af_instance, expected_mats_per_dag)


def test_custom_callback_behavior(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that custom callbacks to proxying_to_dagster are properly applied."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()

    expected_mats_per_dag = {
        "affected_dag": [
            AssetKey("affected_dag__print_asset"),
            AssetKey("affected_dag__another_print_asset"),
        ],
        "unaffected_dag": [
            AssetKey("unaffected_dag__print_asset"),
            AssetKey("unaffected_dag__another_print_asset"),
        ],
    }

    poll_for_expected_mats(af_instance, expected_mats_per_dag)

    for task_id in ["print_task", "downstream_print_task"]:
        affected_print_task = af_instance.get_task_info(dag_id="affected_dag", task_id=task_id)
        assert affected_print_task.metadata["retries"] == 1
        unaffected_print_task = af_instance.get_task_info(dag_id="unaffected_dag", task_id=task_id)
        assert unaffected_print_task.metadata["retries"] == 0


def test_migrated_overridden_dag_custom_operator_materializes(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that assets are properly materialized from an overridden dag, and that the proxied task retains attributes from the custom operator."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()
    assert af_instance.get_task_info(dag_id="overridden_dag_custom_callback", task_id="OVERRIDDEN")

    expected_mats_per_dag = {
        "overridden_dag_custom_callback": [AssetKey("asset_overridden_dag_custom_callback")],
    }
    poll_for_expected_mats(af_instance, expected_mats_per_dag)


def test_partitioned_observation(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that assets with time-window partitions get partitions mapped correctly onto their materializations."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()
    af_run_id = af_instance.trigger_dag(
        dag_id="daily_interval_dag", logical_date=get_current_datetime_midnight()
    )
    af_instance.wait_for_run_completion(dag_id="daily_interval_dag", run_id=af_run_id, timeout=30)

    dagster_instance = DagsterInstance.get()
    entry = poll_for_materialization(
        dagster_instance=dagster_instance,
        asset_key=AssetKey("daily_interval_dag__partitioned"),
    )
    assert entry.asset_materialization
    assert entry.asset_materialization.partition


def test_assets_multiple_jobs_same_task(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test the case where multiple assets within the same task have different jobs. Ensure we still materialize them correctly."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()
    assert af_instance.get_task_info(dag_id="overridden_dag_custom_callback", task_id="OVERRIDDEN")

    expected_mats_per_dag = {
        "multi_job_assets_dag": [
            AssetKey("multi_job__a"),
            AssetKey("multi_job__b"),
            AssetKey("multi_job__c"),
        ],
    }
    poll_for_expected_mats(af_instance, expected_mats_per_dag)


def test_partitioned_migrated(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that partitioned assets are properly materialized from a proxied task."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()
    af_instance.unpause_dag(dag_id="migrated_daily_interval_dag")
    # Wait for dag run to exist
    expected_logical_date = get_current_datetime_midnight() - timedelta(days=1)
    expected_run_id = f"scheduled__{expected_logical_date.isoformat()}"
    poll_for_airflow_run_existence_and_completion(
        af_instance=af_instance,
        dag_id="migrated_daily_interval_dag",
        af_run_id=expected_run_id,
        duration=30,
    )
    dagster_instance = DagsterInstance.get()
    entry = poll_for_materialization(
        dagster_instance=dagster_instance,
        asset_key=AssetKey("migrated_daily_interval_dag__partitioned"),
    )
    assert entry.asset_materialization
    assert entry.asset_materialization.partition
    assert entry.asset_materialization.partition == expected_logical_date.strftime("%Y-%m-%d")


def test_success_on_retry(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that success upon retrying an asset means that the launched airflow run succeeds."""
    from kitchen_sink.airflow_instance import local_airflow_instance
    from kitchen_sink.dagster_defs.retries_configured import succeeds_on_final_retry

    af_instance = local_airflow_instance()
    assert af_instance.get_task_info(dag_id="migrated_asset_has_retries", task_id="my_task")

    expected_mats_per_dag = {
        "migrated_asset_has_retries": [succeeds_on_final_retry.key],
    }
    poll_for_expected_mats(af_instance, expected_mats_per_dag)


def test_failure_when_asset_failures_tag_set(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that when an asset fails and the run has the "RETRY_ON_ASSET_OR_OP_FAILURE_TAG" set to False,
    that the dag run fails.
    """
    dag_id = "migrated_asset_has_retries_not_step_failure"
    af_instance = local_airflow_instance()
    assert af_instance.get_task_info(dag_id=dag_id, task_id="my_task")

    run_id = af_instance.trigger_dag(dag_id=dag_id)
    af_instance.wait_for_run_completion(dag_id=dag_id, run_id=run_id)
    assert af_instance.get_run_state(dag_id=dag_id, run_id=run_id) == "failed"

    with DagsterInstance.get() as dg_instance:
        # There should be a single run launched from the task, and it should be failed.
        runs = dg_instance.get_runs(filters=RunsFilter(tags={DAG_RUN_ID_TAG_KEY: run_id}))
        assert len(runs) == 1
        assert next(iter(runs)).status == DagsterRunStatus.FAILURE


def test_respect_airflow_retries(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Airflow doesn't actually have dag-level retries; only task-level retries. This test just ensures that we handle task-level retries gracefully."""
    from kitchen_sink.dagster_defs.mapped_defs import UNMAPPED_SPECS_INSTANCE_NAME, materialize_dags

    af_instance = local_airflow_instance(UNMAPPED_SPECS_INSTANCE_NAME)

    dagster_instance = DagsterInstance.get()
    result = materialize(
        assets=[materialize_dags],
        instance=dagster_instance,
        selection=AssetSelection.keys([af_instance.name, "dag", "unmapped__dag_with_retries"]),
    )
    assert result.success



from datetime import timedelta
from pathlib import Path

from airflow import DAG
from airflow.operators.python import PythonOperator
from dagster._time import get_current_datetime_midnight
from dagster_airlift.in_airflow import proxying_to_dagster
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml


def print_hello() -> None:
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "retries": 0,
}

with DAG(
    dag_id="migrated_daily_interval_dag",
    default_args=default_args,
    schedule="@daily",
    start_date=get_current_datetime_midnight() - timedelta(days=1),
    # We pause this dag upon creation to avoid running it immediately
    is_paused_upon_creation=True,
) as minute_dag:
    PythonOperator(task_id="my_task", python_callable=print_hello)


proxying_to_dagster(
    proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
    global_vars=globals(),
)



from dagster import AssetKey, AssetOut, Definitions, asset, multi_asset


@asset
def some_asset():
    return "asset_value"


@asset
def other_asset():
    return "other_asset_value"


@asset(key=["nested", "asset"])
def nested_asset():
    return "nested_asset_value"


@asset(key=["string", "interpretation"])
def string_interpretation():
    return "string_interpretation_value"


@multi_asset(outs={"my_out": AssetOut(key=AssetKey(["backslash/interpretation"]))})
def backslash_interpretation():
    return "backslash_interpretation_value"


defs = Definitions(
    assets=[some_asset, other_asset, nested_asset, string_interpretation, backslash_interpretation]
)



from dagster._core.definitions.asset_key import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster_airlift.constants import TASK_MAPPING_METADATA_KEY
from dagster_airlift.core.airflow_instance import DagInfo
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.serialization.compute import (
    FetchedAirflowData,
    TaskHandle,
    build_airlift_metadata_mapping_info,
    fetch_all_airflow_data,
)
from dagster_airlift.core.serialization.serialized_data import TaskInfo
from dagster_airlift.core.utils import metadata_for_task_mapping
from dagster_airlift.test import AirflowInstanceFake


def ak(key: str) -> AssetKey:
    return AssetKey.from_user_string(key)


def airlift_asset_spec(key: CoercibleToAssetKey, dag_id: str, task_id: str) -> AssetSpec:
    return AssetSpec(key=key, metadata=metadata_for_task_mapping(task_id=task_id, dag_id=dag_id))


def airlift_multiple_task_asset_spec(
    key: CoercibleToAssetKey, handles: list[TaskHandle]
) -> AssetSpec:
    return AssetSpec(
        key=key, metadata={TASK_MAPPING_METADATA_KEY: [handle._asdict() for handle in handles]}
    )


def test_build_task_mapping_info_no_mapping() -> None:
    spec_mapping_info = build_airlift_metadata_mapping_info(
        mapped_assets=[AssetSpec("asset1"), AssetSpec("asset2")]
    )
    assert len(spec_mapping_info.dag_ids) == 0
    assert not (spec_mapping_info.asset_keys_by_mapped_task_id)
    assert not (spec_mapping_info.task_handle_map)


def test_build_single_task_spec() -> None:
    spec_mapping_info = build_airlift_metadata_mapping_info(
        mapped_assets=[airlift_asset_spec("asset1", "dag1", "task1")]
    )
    assert spec_mapping_info.dag_ids == {"dag1"}
    assert spec_mapping_info.task_id_map == {"dag1": {"task1"}}
    assert spec_mapping_info.all_mapped_asset_keys_by_dag_id == {"dag1": {ak("asset1")}}
    assert spec_mapping_info.asset_keys_by_mapped_task_id == {"dag1": {"task1": {ak("asset1")}}}
    assert spec_mapping_info.task_handle_map == {
        ak("asset1"): set([TaskHandle(dag_id="dag1", task_id="task1")])
    }


def test_task_with_multiple_assets() -> None:
    spec_mapping_info = build_airlift_metadata_mapping_info(
        mapped_assets=[
            airlift_asset_spec("asset1", "dag1", "task1"),
            airlift_asset_spec("asset2", "dag1", "task1"),
            airlift_asset_spec("asset3", "dag1", "task1"),
            airlift_asset_spec("asset4", "dag2", "task1"),
        ]
    )

    assert spec_mapping_info.dag_ids == {"dag1", "dag2"}
    assert spec_mapping_info.task_id_map == {"dag1": {"task1"}, "dag2": {"task1"}}
    assert spec_mapping_info.all_mapped_asset_keys_by_dag_id == {
        "dag1": {ak("asset1"), ak("asset2"), ak("asset3")},
        "dag2": {ak("asset4")},
    }
    assert spec_mapping_info.asset_keys_by_mapped_task_id == {
        "dag1": {"task1": {ak("asset1"), ak("asset2"), ak("asset3")}},
        "dag2": {"task1": {ak("asset4")}},
    }

    assert spec_mapping_info.task_handle_map == {
        ak("asset1"): set([TaskHandle(dag_id="dag1", task_id="task1")]),
        ak("asset2"): set([TaskHandle(dag_id="dag1", task_id="task1")]),
        ak("asset3"): set([TaskHandle(dag_id="dag1", task_id="task1")]),
        ak("asset4"): set([TaskHandle(dag_id="dag2", task_id="task1")]),
    }


def test_map_multiple_tasks_to_single_asset() -> None:
    spec_mapping_info = build_airlift_metadata_mapping_info(
        mapped_assets=[
            airlift_multiple_task_asset_spec(
                "asset1",
                [
                    TaskHandle(dag_id="dag1", task_id="task1"),
                    TaskHandle(dag_id="dag2", task_id="task1"),
                ],
            ),
        ]
    )

    assert spec_mapping_info.dag_ids == {"dag1", "dag2"}
    assert spec_mapping_info.task_id_map == {"dag1": {"task1"}, "dag2": {"task1"}}
    assert spec_mapping_info.all_mapped_asset_keys_by_dag_id == {
        "dag1": {ak("asset1")},
        "dag2": {ak("asset1")},
    }

    assert spec_mapping_info.asset_keys_by_mapped_task_id == {
        "dag1": {"task1": {ak("asset1")}},
        "dag2": {"task1": {ak("asset1")}},
    }

    assert spec_mapping_info.task_handle_map == {
        ak("asset1"): set(
            [TaskHandle(dag_id="dag1", task_id="task1"), TaskHandle(dag_id="dag2", task_id="task1")]
        )
    }


def test_fetched_airflow_data() -> None:
    ws_url = "http://localhost:8080"

    fetched_airflow_data = FetchedAirflowData(
        dag_infos={
            "dag1": DagInfo(
                webserver_url=ws_url,
                dag_id="dag1",
                metadata={},
            )
        },
        task_info_map={
            "dag1": {
                "task1": TaskInfo(
                    webserver_url=ws_url, dag_id="dag1", task_id="task1", metadata={}
                ),
                "task2": TaskInfo(
                    webserver_url=ws_url, dag_id="dag1", task_id="task2", metadata={}
                ),
            }
        },
        mapping_info=build_airlift_metadata_mapping_info(
            mapped_assets=[
                airlift_asset_spec("asset1", "dag1", "task1"),
                airlift_asset_spec("asset2", "dag1", "task2"),
            ]
        ),
        datasets=[],
    )

    all_mapped_tasks = fetched_airflow_data.all_mapped_tasks
    assert all_mapped_tasks.keys() == {ak("asset1"), ak("asset2")}
    assert all_mapped_tasks[ak("asset1")] == {TaskHandle(dag_id="dag1", task_id="task1")}


def test_produce_fetched_airflow_data() -> None:
    mapping_info = build_airlift_metadata_mapping_info(
        mapped_assets=[
            airlift_asset_spec("asset1", "dag1", "task1"),
            AssetSpec("asset2", deps=[ak("asset1")]),
        ]
    )

    instance = AirflowInstanceFake(
        dag_infos=[
            DagInfo(
                dag_id="dag1",
                metadata={"file_token": "dummy_file_token"},
                webserver_url="http://localhost:8080",
            )
        ],
        task_infos=[
            TaskInfo(
                dag_id="dag1",
                task_id="task1",
                metadata={},
                webserver_url="http://localhost:8080",
            )
        ],
        task_instances=[],
        dag_runs=[],
    )

    fetched_airflow_data = fetch_all_airflow_data(
        airflow_instance=instance,
        mapping_info=mapping_info,
        dag_selector_fn=None,
        automapping_enabled=True,
        retrieval_filter=AirflowFilter(),
    )

    assert len(fetched_airflow_data.mapping_info.mapped_task_asset_specs) == 1
    assert len(list(fetched_airflow_data.mapping_info.asset_specs)) == 2
    assert fetched_airflow_data.mapping_info.downstream_deps == {ak("asset1"): {ak("asset2")}}



from datetime import datetime

from airflow import DAG
from airflow.datasets import Dataset
from airflow.operators.python import PythonOperator

default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 1,
}


def print_fn() -> None:
    import os

    os.environ["NO_PROXY"] = "*"
    import json

    print("Hello")  # noqa: T201
    data = json.dumps(
        {
            "foo": "bar",
            "my_timestamp": {"raw_value": 111, "type": "timestamp"},
        }
    )
    print(f"DAGSTER_START{data}DAGSTER_END")  # noqa: T201
    another_data = json.dumps(
        {
            "foo": "baz",
            "my_other_timestamp": {"raw_value": 113, "type": "timestamp"},
        }
    )
    print(f"DAGSTER_START{another_data}DAGSTER_END")  # noqa: T201


# Inter-dag structure as follows:
# dataset_producer -> example1_consumer
#                  -> example2_consumer

example1 = Dataset("s3://dataset-bucket/example1.csv")
example2 = Dataset("s3://dataset-bucket/example2.csv")
with DAG(
    "dataset_producer",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as dataset_producer_dag:
    print_task = PythonOperator(
        task_id="print_task", python_callable=print_fn, outlets=[example1, example2]
    )
with DAG(
    "example1_consumer",
    start_date=datetime(2023, 1, 1),
    schedule=[example1],
) as example1_consumer_dag:
    print_task = PythonOperator(task_id="print_task", python_callable=print_fn)

with DAG(
    "example2_consumer",
    start_date=datetime(2023, 1, 1),
    schedule=[example2],
) as example2_consumer_dag:
    print_task = PythonOperator(task_id="print_task", python_callable=print_fn)



from dagster import AssetsDefinition, Definitions
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster_airlift.core import (
    AirflowBasicAuthBackend,
    AirflowInstance,
    build_defs_from_airflow_instance,
)
from dagster_airlift.core.top_level_dag_def_api import assets_with_task_mappings

from perf_harness.dagster_defs.constants import (
    AIRFLOW_BASE_URL,
    AIRFLOW_INSTANCE_NAME,
    PASSWORD,
    USERNAME,
)
from perf_harness.shared.constants import get_num_assets, get_num_dags, get_num_tasks

airflow_instance = AirflowInstance(
    auth_backend=AirflowBasicAuthBackend(
        webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
    ),
    name=AIRFLOW_INSTANCE_NAME,
)


def build_asset(specs: list[AssetSpec]) -> AssetsDefinition:
    @multi_asset(specs=specs)
    def asset_fn(_):
        pass

    return asset_fn


def build_asset_for_task(task_name: str, prev_asset_specs: list[AssetSpec]) -> AssetsDefinition:
    specs = [
        # Create a bunch of dependencies for each asset.
        AssetSpec(f"{task_name}_asset_{i}", deps=[spec.key for spec in prev_asset_specs])
        for i in range(get_num_assets())
    ]
    return build_asset(specs)


def get_dag_defs() -> Definitions:
    all_dag_defs = []
    prev_asset_specs: list[AssetSpec] = []
    for i in range(get_num_dags()):
        dag_id = f"dag_{i}"
        task_mappings = {}
        for j in range(get_num_tasks()):
            task_name = f"task_{i}_{j}"
            asset = build_asset_for_task(task_name, prev_asset_specs)
            task_mappings[task_name] = [asset]
            prev_asset_specs = asset.specs  # type: ignore
        all_dag_defs.append(
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id=dag_id,
                    task_mappings=task_mappings,
                )
            )
        )
    return Definitions.merge(*all_dag_defs)


defs = build_defs_from_airflow_instance(airflow_instance=airflow_instance, defs=get_dag_defs())



import os
import time
from datetime import timedelta

import pytest
from dagster import AssetKey, DagsterInstance
from dagster._core.definitions.metadata.metadata_value import JsonMetadataValue
from dagster._core.events.log import EventLogEntry
from dagster._time import get_current_datetime

from kitchen_sink_tests.integration_tests.conftest import makefile_dir

RAW_METADATA_KEY = "Run Metadata (raw)"


def dag_id_of_mat(event_log_entry: EventLogEntry) -> bool:
    assert event_log_entry.asset_materialization
    assert isinstance(event_log_entry.asset_materialization.metadata, dict)
    json_metadata_value = event_log_entry.asset_materialization.metadata[RAW_METADATA_KEY]
    assert isinstance(json_metadata_value, JsonMetadataValue)
    assert isinstance(json_metadata_value.data, dict)
    return json_metadata_value.data["dag_id"]


def poll_for_observation(
    dagster_instance: DagsterInstance,
    asset_key: AssetKey,
) -> EventLogEntry:
    start_time = get_current_datetime()
    while get_current_datetime() - start_time < timedelta(seconds=30):
        records = dagster_instance.fetch_observations(records_filter=asset_key, limit=1).records
        if len(records) > 0:
            return records[0].event_log_entry
        time.sleep(0.1)

    raise Exception(f"Timeout waiting for observation event on {asset_key}")


@pytest.fixture(name="dagster_home")
def dagster_home_fixture(local_env: None) -> str:
    return os.environ["DAGSTER_HOME"]


@pytest.fixture(name="dagster_dev_cmd")
def dagster_dev_cmd_fixture() -> list[str]:
    return ["make", "run_observation_defs", "-C", str(makefile_dir())]


def test_observation_defs_are_observed(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that assets can load properly, and that observations register."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()
    dag_ids = [dag_info.dag_id for dag_info in af_instance.list_dags()]
    assert "simple_unproxied_dag" in dag_ids

    expected_obs_per_dag = {
        "simple_unproxied_dag": [AssetKey("my_asset"), AssetKey("my_downstream_asset")],
    }

    for dag_id, expected_asset_keys in expected_obs_per_dag.items():
        airflow_run_id = af_instance.trigger_dag(dag_id=dag_id)
        af_instance.wait_for_run_completion(dag_id=dag_id, run_id=airflow_run_id, timeout=120)
        dagster_instance = DagsterInstance.get()

        dag_asset_key = AssetKey(["my_airflow_instance", "dag", dag_id])
        assert poll_for_observation(dagster_instance, dag_asset_key)

        for expected_asset_key in expected_asset_keys:
            obs_entry = poll_for_observation(dagster_instance, expected_asset_key)
            assert obs_entry.asset_observation
            assert obs_entry.asset_observation.asset_key == expected_asset_key



from datetime import timedelta

from dagster import Definitions, asset, define_asset_job, multi_asset
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.partitions.definition import DailyPartitionsDefinition
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._time import get_current_datetime_midnight
from dagster_airlift.core import (
    assets_with_dag_mappings,
    assets_with_task_mappings,
    build_defs_from_airflow_instance,
    load_airflow_dag_asset_specs,
)
from dagster_airlift.core.multiple_tasks import targeted_by_multiple_tasks

from kitchen_sink.airflow_instance import local_airflow_instance
from kitchen_sink.dagster_defs.retries_configured import just_fails, succeeds_on_final_retry


def make_print_asset(key: str) -> AssetsDefinition:
    @asset(key=key)
    def print_asset() -> None:
        # ruff: noqa: T201
        print("Hello, world!")

    return print_asset


@asset(description="Asset one is materialized by multiple airflow tasks")
def asset_one() -> None:
    # ruff: noqa: T201
    print("Materialized asset one")


@asset(description="Asset two is materialized by an overridden dag")
def asset_two() -> None:
    print("Materialized asset two")


@asset(description="Materialized by overridden_dag_custom_callback")
def asset_overridden_dag_custom_callback() -> None:
    print("Materialized by overridden_dag_custom_callback")


# Assets within multi_job_assets_dag
@asset
def multi_job__a() -> None:
    print("Materialized a")


@asset(deps=[multi_job__a])
def multi_job__b() -> None:
    print("Materialized b")


@asset(deps=[multi_job__a])
def multi_job__c() -> None:
    print("Materialized c")


job1 = define_asset_job("job1", [multi_job__a])
job2 = define_asset_job("job2", [multi_job__b, multi_job__c])


# Partitioned assets for migrated_daily_interval_dag
@asset(
    partitions_def=DailyPartitionsDefinition(
        start_date=get_current_datetime_midnight() - timedelta(days=2)
    )
)
def migrated_daily_interval_dag__partitioned() -> None:
    print("Materialized daily_interval_dag__partitioned")


def build_mapped_defs() -> Definitions:
    return build_defs_from_airflow_instance(
        airflow_instance=local_airflow_instance(),
        dag_selector_fn=lambda dag: not dag.dag_id.startswith("unmapped"),
        sensor_minimum_interval_seconds=1,
        defs=Definitions.merge(
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="print_dag",
                    task_mappings={
                        "print_task": [make_print_asset("print_asset")],
                        "downstream_print_task": [make_print_asset("another_print_asset")],
                    },
                ),
            ),
            targeted_by_multiple_tasks(
                Definitions([asset_one]),
                task_handles=[
                    {"dag_id": "weekly_dag", "task_id": "asset_one_weekly"},
                    {"dag_id": "daily_dag", "task_id": "asset_one_daily"},
                ],
            ),
            Definitions(assets=assets_with_dag_mappings({"overridden_dag": [asset_two]})),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="affected_dag",
                    task_mappings={
                        "print_task": [make_print_asset("affected_dag__print_asset")],
                        "downstream_print_task": [
                            make_print_asset("affected_dag__another_print_asset")
                        ],
                    },
                ),
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="unaffected_dag",
                    task_mappings={
                        "print_task": [make_print_asset("unaffected_dag__print_asset")],
                        "downstream_print_task": [
                            make_print_asset("unaffected_dag__another_print_asset")
                        ],
                    },
                ),
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="daily_interval_dag",
                    task_mappings={
                        "task": [
                            AssetSpec(
                                key="daily_interval_dag__partitioned",
                                partitions_def=DailyPartitionsDefinition(
                                    start_date=get_current_datetime_midnight()
                                ),
                            )
                        ],
                    },
                ),
            ),
            Definitions(
                assets_with_dag_mappings(
                    {"overridden_dag_custom_callback": [asset_overridden_dag_custom_callback]}
                )
            ),
            Definitions(
                assets_with_task_mappings(
                    dag_id="multi_job_assets_dag",
                    task_mappings={"print_task": [multi_job__a, multi_job__b, multi_job__c]},
                ),
                jobs=[job1, job2],
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="migrated_daily_interval_dag",
                    task_mappings={"my_task": [migrated_daily_interval_dag__partitioned]},
                ),
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="migrated_asset_has_retries",
                    task_mappings={"my_task": [succeeds_on_final_retry]},
                )
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="migrated_asset_has_retries_not_step_failure",
                    task_mappings={"my_task": [just_fails]},
                )
            ),
        ),
    )


# We need to use a different name for this instance because the reconstruction metadata key is
# derived from the instance name, and we don't want the reconstruction metadata from the mapped defs
# instance to clobber the reconstruction metadata from this.
UNMAPPED_SPECS_INSTANCE_NAME = "unmapped_specs_instance"

unmapped_specs = load_airflow_dag_asset_specs(
    airflow_instance=local_airflow_instance(UNMAPPED_SPECS_INSTANCE_NAME),
    dag_selector_fn=lambda dag: dag.dag_id.startswith("unmapped"),
)


@multi_asset(specs=unmapped_specs)
def materialize_dags(context: AssetExecutionContext):
    for spec in unmapped_specs:
        af_instance = local_airflow_instance()
        dag_id = spec.metadata["Dag ID"]
        dag_run_id = af_instance.trigger_dag(dag_id=dag_id)
        af_instance.wait_for_run_completion(dag_id=dag_id, run_id=dag_run_id)
        state = af_instance.get_run_state(dag_id=dag_id, run_id=dag_run_id)
        if state != "success":
            raise Exception(f"Failed to materialize {dag_id} with state {state}")


defs = Definitions.merge(build_mapped_defs(), Definitions([materialize_dags]))



from datetime import datetime

from airflow import DAG
from airflow.operators.python import PythonOperator


def print_hello():
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 1,
}

dag = DAG(
    "print_dag", default_args=default_args, schedule_interval=None, is_paused_upon_creation=False
)
print_op = PythonOperator(task_id="print_task", python_callable=print_hello, dag=dag)
downstream_print_op = PythonOperator(
    task_id="downstream_print_task", python_callable=print_hello, dag=dag
)

downstream_print_op.set_upstream(print_op)



from dagster import Definitions
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster_airlift.core import assets_with_task_mappings
from dagster_airlift.core.load_defs import build_job_based_airflow_defs

from kitchen_sink.airflow_instance import local_airflow_instance


def build_job_based_defs() -> Definitions:
    @asset(key="print_asset")
    def print_asset(context: AssetExecutionContext) -> None:
        context.add_output_metadata({"foo": "bar"})

    @asset(key="another_print_asset")
    def another_print_asset(context: AssetExecutionContext) -> None:
        context.add_output_metadata({"foo": "baz"})

    return build_job_based_airflow_defs(
        airflow_instance=local_airflow_instance(),
        mapped_defs=Definitions(
            assets=[
                *assets_with_task_mappings(
                    dag_id="deferred_events_dag",
                    task_mappings={
                        "print_task": [print_asset],
                        "downstream_print_task": [another_print_asset],
                    },
                ),
            ]
        ),
    )


defs = build_job_based_defs()



from unittest import mock

import boto3
from dagster_airlift.mwaa import MwaaSessionAuthBackend


def test_mwaa_session_auth_from_profile() -> None:
    """Test output format of mwaa auth backend. Does not actually make any live requests to boto or exercise usage of aws apis in any way."""
    with mock.patch("dagster_airlift.mwaa.auth.get_session_info") as mock_get_session_info:
        mock_get_session_info.return_value = ("my-webserver-hostname", "my-session-cookie")
        auth_backend = MwaaSessionAuthBackend.from_profile(region="us-west-2", env_name="my-env")
        session = auth_backend.get_session()
        assert session.cookies["session"] == "my-session-cookie"
        assert auth_backend.get_webserver_url() == "https://my-webserver-hostname"


def test_mwaa_session_auth_direct_mwaa_client_creation() -> None:
    """Test output format of mwaa auth backend. Does not actually make any live requests to boto or exercise usage of aws apis in any way."""
    with mock.patch("dagster_airlift.mwaa.auth.get_session_info") as mock_get_session_info:
        mock_get_session_info.return_value = ("my-webserver-hostname", "my-session-cookie")
        boto_session = boto3.Session(region_name="us-west-2")
        mwaa = boto_session.client("mwaa")
        auth_backend = MwaaSessionAuthBackend(mwaa_client=mwaa, env_name="my-env")
        session = auth_backend.get_session()
        assert session.cookies["session"] == "my-session-cookie"
        assert auth_backend.get_webserver_url() == "https://my-webserver-hostname"



import os

import pytest
from dagster_airlift.test.test_utils import get_job_from_defs


@pytest.mark.skipif(
    os.getenv("BUILDKITE") is not None,
    reason="Skipping MWAA tests on Buildkite. We need to add creds to buildkite which can access the MWAA environment.",
)
def test_mwaa_backend():
    from dagster_airlift_tests.mwaa_integration_tests.dagster_defs.defs import defs_obj

    assert defs_obj.jobs
    assert len(list(defs_obj.jobs)) > 1
    assert get_job_from_defs("dag_0", defs_obj)
    assert get_job_from_defs("my_airflow_instance__airflow_monitoring_job", defs_obj)



import copy

import pytest
from airflow.operators.python import PythonOperator
from dagster_airlift.in_airflow import proxying_to_dagster
from dagster_airlift.in_airflow.base_asset_operator import BaseDagsterAssetsOperator
from dagster_airlift.in_airflow.proxied_state import AirflowProxiedState

from dagster_airlift_tests.unit_tests.in_airflow_tests.conftest import (
    build_dags_dict_given_structure,
)


def test_proxying_to_dagster() -> None:
    """Test that we can proxy a set of dags, and as a result, operators are replaced and tags are added."""
    globals_fake = build_dags_dict_given_structure(
        {
            "task_is_proxied": {"task": []},
            "initially_not_proxied": {"task": []},
            "should_be_ignored": {"task": []},
        }
    )
    original_globals = copy.deepcopy(globals_fake)
    proxying_to_dagster(
        global_vars=globals_fake,
        proxied_state=AirflowProxiedState.from_dict(
            {
                "task_is_proxied": {"tasks": [{"id": "task", "proxied": True}]},
                "initially_not_proxied": {"tasks": [{"id": "task", "proxied": False}]},
            }
        ),
    )
    # Both dags should have the tag "Proxied to Dagster"
    assert len(globals_fake["task_is_proxied"].tags) == 1
    assert len(globals_fake["initially_not_proxied"].tags) == 1

    assert "1 Task Marked as Proxied to Dagster" in globals_fake["task_is_proxied"].tags
    assert "0 Tasks Marked as Proxied to Dagster" in globals_fake["initially_not_proxied"].tags

    # Only the task marked as proxied should be replaced with a DagsterOperator
    assert isinstance(globals_fake["task_is_proxied"].task_dict["task"], BaseDagsterAssetsOperator)
    assert isinstance(globals_fake["initially_not_proxied"].task_dict["task"], PythonOperator)
    assert isinstance(globals_fake["should_be_ignored"].task_dict["task"], PythonOperator)

    # Change initially_not_proxied to be proxied
    proxying_to_dagster(
        global_vars=original_globals,
        proxied_state=AirflowProxiedState.from_dict(
            {
                "task_is_proxied": {"tasks": [{"id": "task", "proxied": True}]},
                "initially_not_proxied": {"tasks": [{"id": "task", "proxied": True}]},
            }
        ),
    )
    # Both dags should have the tag "Proxied to Dagster"
    assert len(original_globals["task_is_proxied"].tags) == 1
    assert len(original_globals["initially_not_proxied"].tags) == 1

    assert "1 Task Marked as Proxied to Dagster" in original_globals["task_is_proxied"].tags
    assert "1 Task Marked as Proxied to Dagster" in original_globals["initially_not_proxied"].tags

    # Both tasks should be replaced with a DagsterOperator
    assert isinstance(
        original_globals["task_is_proxied"].task_dict["task"], BaseDagsterAssetsOperator
    )
    assert isinstance(
        original_globals["initially_not_proxied"].task_dict["task"],
        BaseDagsterAssetsOperator,
    )
    assert isinstance(original_globals["should_be_ignored"].task_dict["task"], PythonOperator)


def test_proxying_to_dagster_no_dags() -> None:
    """Ensure that we error when no dags are found in the current context."""
    with pytest.raises(Exception, match=r"No dags found in globals dictionary."):
        proxying_to_dagster(global_vars={}, proxied_state=AirflowProxiedState.from_dict({}))


def test_proxying_to_dagster_dags_dont_exist() -> None:
    """Ensure that we dont error when a dag is referenced that does not exist."""
    globals_fake = build_dags_dict_given_structure({"dag": {"task": []}})
    proxying_to_dagster(
        global_vars=globals_fake,
        proxied_state=AirflowProxiedState.from_dict(
            {
                "doesnt_exist": {"tasks": [{"id": "task", "proxied": True}]},
            }
        ),
    )
    dag = globals_fake["dag"]
    assert len(dag.tags) == 0


def test_proxying_to_dagster_task_doesnt_exist() -> None:
    """Ensure that we error when a task is referenced that does not exist."""
    globals_fake = build_dags_dict_given_structure({"dag": {"task": []}})
    with pytest.raises(Exception, match="Task with id `doesnt_exist` not found in dag `dag`"):
        proxying_to_dagster(
            global_vars=globals_fake,
            proxied_state=AirflowProxiedState.from_dict(
                {
                    "dag": {"tasks": [{"id": "doesnt_exist", "proxied": True}]},
                }
            ),
        )



from dagster._core.definitions.utils import check_valid_name
from dagster_airlift.core.airflow_instance import DagInfo
from dagster_airlift.core.serialization.defs_construction import make_default_dag_asset_key

INSTANCE_NAME = "airflow_instance"


def produce_valid_name(dag_info: DagInfo) -> str:
    return (
        make_default_dag_asset_key("airflow_instance", dag_info.dag_id)
        .to_user_string()
        .replace("/", "__")
    )


def test_dag_info_asset_key() -> None:
    # Airflow allows "exclusively alphanumeric characters, dashes, dots and underscores (all ASCII)"
    dag_info_hyphen = DagInfo(webserver_url="", dag_id="my-test-dag-id", metadata={})
    assert check_valid_name(produce_valid_name(dag_info_hyphen))

    dag_info_slash = DagInfo(webserver_url="", dag_id="my-test/dag-id", metadata={})
    assert check_valid_name(produce_valid_name(dag_info_slash))

    dag_info_dot = DagInfo(webserver_url="", dag_id="my-test.dag-id", metadata={})
    assert check_valid_name(produce_valid_name(dag_info_dot))



import json
import tempfile
from collections.abc import Iterator
from pathlib import Path
from unittest import mock

import dagster_airlift.core as dg_airlift_core
import pytest
import yaml
from click.testing import CliRunner
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.instance_for_test import instance_for_test
from dagster._core.storage.defs_state.base import DefsStateStorage
from dagster._core.test_utils import ensure_dagster_tests_import
from dagster._utils import pushd
from dagster._utils.test.definitions import scoped_definitions_load_context
from dagster_airlift.constants import (
    DAG_MAPPING_METADATA_KEY,
    SOURCE_CODE_METADATA_KEY,
    TASK_MAPPING_METADATA_KEY,
)
from dagster_airlift.core.components.airflow_instance.component import AirflowInstanceComponent
from dagster_airlift.test import make_instance
from dagster_airlift.test.test_utils import asset_spec, get_job_from_defs
from dagster_dg_cli.cli import cli

ensure_dagster_tests_import()
from dagster_tests.components_tests.utils import (
    build_component_defs_for_test,
    temp_code_location_bar,
)


@pytest.fixture(autouse=True)
def _setup() -> Iterator:
    with instance_for_test() as instance, scoped_definitions_load_context():
        yield instance


@pytest.fixture
def temp_cwd() -> Iterator[Path]:
    with tempfile.TemporaryDirectory() as temp_dir:
        with pushd(temp_dir):
            yield Path.cwd()


@pytest.fixture
def component_for_test(temp_cwd: Path):
    airflow_instance = make_instance(
        {"dag_1": ["dag_1_task_1", "dag_1_task_2"], "dag_2": ["dag_2_task_1", "dag_2_task_2"]},
        dataset_construction_info=[
            {
                "uri": "s3://dataset-bucket/example1.csv",
                "producing_tasks": [
                    {"dag_id": "dag_1", "task_id": "dag_1_task_1"},
                ],
                "consuming_dags": ["dag_2"],
            },
            {
                "uri": "s3://dataset-bucket/example2.csv",
                "producing_tasks": [
                    {"dag_id": "dag_2", "task_id": "dag_2_task_1"},
                ],
                "consuming_dags": [],
            },
        ],
    )

    class DebugAirflowInstanceComponent(AirflowInstanceComponent):
        def _get_instance(self) -> dg_airlift_core.AirflowInstance:
            return airflow_instance

    return DebugAirflowInstanceComponent


def test_load_dags_basic(component_for_test: type[AirflowInstanceComponent]) -> None:
    state_storage = DefsStateStorage.get()
    assert state_storage is not None
    defs = build_component_defs_for_test(
        component_type=component_for_test,
        attrs={
            "auth": {
                "type": "basic_auth",
                "webserver_url": "http://localhost:8080",
                "username": "admin",
                "password": "admin",
            },
            "name": "test_instance",
        },
        post_processing={
            "assets": [
                {
                    "target": "*",
                    "attributes": {
                        "metadata": {
                            "foo": "bar",
                        },
                    },
                }
            ]
        },
    )

    for asset_key in ["example1", "example2"]:
        keyed_spec = asset_spec(asset_key, defs)
        assert keyed_spec is not None
        assert keyed_spec.metadata["foo"] == "bar"

    assert defs.jobs
    assert len(list(defs.jobs)) == 3  # monitoring job + 2 dag jobs.
    for job_name in ["dag_1", "dag_2"]:
        job = get_job_from_defs(job_name, defs)
        assert job is not None
        assert isinstance(job, JobDefinition)
        assert job.metadata.get(SOURCE_CODE_METADATA_KEY) is not None


def _scaffold_airlift(scaffold_format: str):
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "scaffold",
            "defs",
            "dagster_airlift.core.components.AirflowInstanceComponent",
            "qux",
            "--json-params",
            json.dumps({"name": "qux", "auth_type": "basic_auth"}),
            "--format",
            scaffold_format,
        ],
    )
    assert result.exit_code == 0


def test_scaffold_airlift_yaml():
    with temp_code_location_bar():
        _scaffold_airlift("yaml")
        assert Path("bar/defs/qux/defs.yaml").exists()
        with open("bar/defs/qux/defs.yaml") as f:
            assert yaml.safe_load(f) == {
                "type": "dagster_airlift.core.components.AirflowInstanceComponent",
                "attributes": {
                    "name": "qux",
                    "auth": {
                        "type": "basic_auth",
                        "webserver_url": '{{ env("AIRFLOW_WEBSERVER_URL") }}',
                        "username": '{{ env("AIRFLOW_USERNAME") }}',
                        "password": '{{ env("AIRFLOW_PASSWORD") }}',
                    },
                },
            }


def test_scaffold_airlift_python():
    with temp_code_location_bar():
        _scaffold_airlift("python")
        assert Path("bar/defs/qux/component.py").exists()
        with open("bar/defs/qux/component.py") as f:
            file_contents = f.read()
            assert file_contents == (
                """import dagster as dg
from dagster_airlift.core.components import AirflowInstanceComponent

@dg.component_instance
def load(context: dg.ComponentLoadContext) -> AirflowInstanceComponent: ...
"""
            )


@pytest.mark.skip("Figure out how to model this test with new tree system")
def test_mapped_assets(component_for_test: type[AirflowInstanceComponent], temp_cwd: Path):
    # Add a sub-dir with an asset that will be task mapped.
    (temp_cwd / "my_asset").mkdir()
    with open(temp_cwd / "my_asset" / "defs.yaml", "w") as f:
        f.write(
            yaml.dump(
                {
                    "type": "dagster_airlift.test.test_utils.BasicAssetComponent",
                    "attributes": {
                        "key": "my_asset",
                    },
                }
            )
        )

    # Add a sub-dir with an asset that will be dag mapped.
    (temp_cwd / "my_asset_2").mkdir()
    with open(temp_cwd / "my_asset_2" / "defs.yaml", "w") as f:
        f.write(
            yaml.dump(
                {
                    "type": "dagster_airlift.test.test_utils.BasicAssetComponent",
                    "attributes": {
                        "key": "my_asset_2",
                    },
                }
            )
        )

    # Next, add an airlift component which references the asset
    defs = build_component_defs_for_test(
        component_type=component_for_test,
        attrs={
            "auth": {
                "type": "basic_auth",
                "webserver_url": "http://localhost:8080",
                "username": "admin",
                "password": "admin",
            },
            "name": "test_instance",
            "mappings": [
                {
                    "dag_id": "dag_1",
                    "task_mappings": [
                        {
                            "task_id": "dag_1_task_1",
                            "assets": [{"by_key": "my_asset"}, {"spec": {"key": "my_ext_asset_1"}}],
                        }
                    ],
                    "assets": [{"by_key": "my_asset_2"}, {"spec": {"key": "my_ext_asset_2"}}],
                }
            ],
        },
        post_processing={
            "assets": [
                {
                    "target": "*",
                    "attributes": {"metadata": {"foo": "bar"}},
                }
            ]
        },
    )

    assert defs.assets is not None
    assert len(list(defs.assets)) == 6

    my_asset_def = next(
        a
        for a in defs.assets
        if isinstance(a, AssetsDefinition) and a.key.to_user_string() == "my_asset"
    )
    my_asset_spec = next(iter(my_asset_def.specs))
    assert my_asset_spec.metadata[TASK_MAPPING_METADATA_KEY] == [
        {
            "dag_id": "dag_1",
            "task_id": "dag_1_task_1",
        }
    ]
    assert DAG_MAPPING_METADATA_KEY not in my_asset_spec.metadata

    my_asset_2_def = next(
        a
        for a in defs.assets
        if isinstance(a, AssetsDefinition) and a.key.to_user_string() == "my_asset_2"
    )
    my_asset_2_spec = next(iter(my_asset_2_def.specs))
    assert my_asset_2_spec.metadata[DAG_MAPPING_METADATA_KEY] == [
        {
            "dag_id": "dag_1",
        }
    ]
    assert TASK_MAPPING_METADATA_KEY not in my_asset_2_spec.metadata

    my_ext_asset_1_spec = next(
        a
        for a in defs.assets
        if isinstance(a, AssetSpec) and a.key.to_user_string() == "my_ext_asset_1"
    )
    assert my_ext_asset_1_spec.metadata[TASK_MAPPING_METADATA_KEY] == [
        {
            "dag_id": "dag_1",
            "task_id": "dag_1_task_1",
        }
    ]
    assert DAG_MAPPING_METADATA_KEY not in my_ext_asset_1_spec.metadata

    my_ext_asset_2_spec = next(
        a
        for a in defs.assets
        if isinstance(a, AssetSpec) and a.key.to_user_string() == "my_ext_asset_2"
    )
    assert my_ext_asset_2_spec.metadata[DAG_MAPPING_METADATA_KEY] == [
        {
            "dag_id": "dag_1",
        }
    ]

    # Datasets should be there too
    assert asset_spec("example1", defs) is not None
    assert asset_spec("example2", defs) is not None


def test_mwaa_auth(component_for_test: type[AirflowInstanceComponent]):
    # mock boto3
    with mock.patch("boto3.Session"):
        defs = build_component_defs_for_test(
            component_for_test,
            {
                "auth": {
                    "type": "mwaa",
                    "env_name": "test_env",
                    "region_name": "test_region",
                    "profile_name": "test_profile",
                },
                "name": "test_instance",
            },
        )
        assert defs.jobs
        assert len(list(defs.jobs)) == 3  # monitoring job + 2 dag jobs.


def test_source_code_retrieval_disabled(
    component_for_test: type[AirflowInstanceComponent], temp_cwd: Path
):
    defs = build_component_defs_for_test(
        component_for_test,
        {
            "auth": {
                "type": "basic_auth",
                "webserver_url": "http://localhost:8080",
                "username": "admin",
                "password": "admin",
            },
            "name": "test_instance",
            "source_code_retrieval_enabled": False,
        },
    )
    assert defs.jobs
    dag1_job = get_job_from_defs("dag_1", defs)
    assert dag1_job is not None
    assert isinstance(dag1_job, JobDefinition)
    assert dag1_job.metadata.get(SOURCE_CODE_METADATA_KEY) is None


def test_airflow_filter(component_for_test: type[AirflowInstanceComponent], temp_cwd: Path):
    defs = build_component_defs_for_test(
        component_for_test,
        {
            "auth": {
                "type": "basic_auth",
                "webserver_url": "http://localhost:8080",
                "username": "admin",
                "password": "admin",
            },
            "name": "test_instance",
            "filter": {
                "dag_id_ilike": "dag_1",
            },
        },
    )
    assert defs.jobs
    dag1_job = get_job_from_defs("dag_1", defs)
    assert dag1_job is not None

    dag2_job = get_job_from_defs("dag_2", defs)
    assert dag2_job is None



import_failed = False
try:
    import dagster as dagster
except ImportError:
    import_failed = True
if not import_failed:
    raise Exception("Importing dagster did not fail")

from dagster_airlift.in_airflow import AirflowProxiedState as AirflowProxiedState



from datetime import datetime

from airflow import DAG
from airflow.operators.python import PythonOperator


def print_hello():
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 1,
}

globals_dict = globals()

for i in range(1000):
    dag = DAG(
        f"dag_{i}", default_args=default_args, schedule_interval=None, is_paused_upon_creation=False
    )
    print_op = PythonOperator(task_id="print_task", python_callable=print_hello, dag=dag)
    downstream_print_op = PythonOperator(
        task_id="downstream_print_task", python_callable=print_hello, dag=dag
    )
    print_op >> downstream_print_op  # pyright: ignore[reportUnusedExpression]
    globals_dict[f"dag_{i}"] = dag



import time
from pathlib import Path

import pytest
import requests
from dagster import AssetKey, DagsterInstance, DagsterRunStatus
from dagster._core.test_utils import environ
from dagster._time import get_current_timestamp
from dagster_airlift.constants import DAG_RUN_ID_TAG_KEY


@pytest.fixture(name="dags_dir")
def dags_dir() -> Path:
    return Path(__file__).parent / "operator_test_project" / "dags"


@pytest.fixture(name="dagster_defs_path")
def dagster_defs_path_fixture() -> str:
    return str(Path(__file__).parent / "operator_test_project" / "dagster_defs.py")


def test_dagster_operator(airflow_instance: None, dagster_dev: None, dagster_home: str) -> None:
    """Tests that dagster operator can correctly map airflow tasks to dagster tasks, and kick off executions."""
    response = requests.post(
        "http://localhost:8080/api/v1/dags/the_dag/dagRuns", auth=("admin", "admin"), json={}
    )
    assert response.status_code == 200, response.json()
    # Wait until the run enters a terminal state
    terminal_status = None
    start_time = get_current_timestamp()
    dag_run = None
    while get_current_timestamp() - start_time < 30:
        response = requests.get(
            "http://localhost:8080/api/v1/dags/the_dag/dagRuns", auth=("admin", "admin")
        )
        assert response.status_code == 200, response.json()
        dag_runs = response.json()["dag_runs"]
        if dag_runs[0]["state"] in ["success", "failed"]:
            terminal_status = dag_runs[0]["state"]
            dag_run = dag_runs[0]
            break
        time.sleep(1)
    assert terminal_status == "success", (
        "Never reached terminal status"
        if terminal_status is None
        else f"terminal status was {terminal_status}"
    )
    with environ({"DAGSTER_HOME": dagster_home}):
        instance = DagsterInstance.get()
        runs = instance.get_runs()
        # The graphql endpoint kicks off a run for each of the tasks in the dag
        assert len(runs) == 2
        some_task_run = [  # noqa
            run
            for run in runs
            if set(list(run.asset_selection)) == {AssetKey(["the_dag__other_task"])}  # type: ignore
        ][0]
        other_task_run = [  # noqa
            run
            for run in runs
            if set(list(run.asset_selection)) == {AssetKey(["the_dag__some_task"])}  # type: ignore
        ][0]
        assert some_task_run.status == DagsterRunStatus.SUCCESS
        assert other_task_run.status == DagsterRunStatus.SUCCESS

        assert isinstance(dag_run, dict)
        assert "dag_run_id" in dag_run
        assert some_task_run.tags[DAG_RUN_ID_TAG_KEY] == dag_run["dag_run_id"]
        assert other_task_run.tags[DAG_RUN_ID_TAG_KEY] == dag_run["dag_run_id"]



from pathlib import Path

import pytest
import yaml
from dagster_airlift.in_airflow.proxied_state import (
    AirflowProxiedState,
    DagProxiedState,
    ProxiedStateParsingError,
    TaskProxiedState,
    load_proxied_state_from_yaml,
)


def test_proxied_state() -> None:
    """Test that we can load proxied state from a yaml file, and that errors are handled in a reasonable way."""
    # First test a valid proxied directory with two files.
    valid_proxied_state_file = Path(__file__).parent / "proxied_state_yamls" / "valid"
    proxied_state = load_proxied_state_from_yaml(valid_proxied_state_file)
    assert isinstance(proxied_state, AirflowProxiedState)
    assert proxied_state == AirflowProxiedState(
        dags={
            "first": DagProxiedState(
                tasks={
                    "first_task": TaskProxiedState(task_id="first_task", proxied=True),
                    "second_task": TaskProxiedState(task_id="second_task", proxied=False),
                    "third_task": TaskProxiedState(task_id="third_task", proxied=True),
                },
                proxied=None,
            ),
            "second": DagProxiedState(
                tasks={
                    "some_task": TaskProxiedState("some_task", proxied=True),
                    "other_task": TaskProxiedState("other_task", proxied=False),
                },
                proxied=None,
            ),
        }
    )

    # Test various incorrect yaml dirs.
    incorrect_dirs = ["empty_file", "nonexistent_dir", "extra_key", "nonsense"]
    for incorrect_dir in incorrect_dirs:
        incorrect_proxied_state_file = Path(__file__).parent / "proxied_state_yamls" / incorrect_dir
        with pytest.raises(ProxiedStateParsingError, match="Error parsing proxied state yaml"):
            load_proxied_state_from_yaml(incorrect_proxied_state_file)


def test_proxied_state_from_yaml() -> None:
    proxied_state_dict = yaml.safe_load("""
tasks:
  - id: load_raw_customers
    proxied: False
  - id: build_dbt_models
    proxied: False
  - id: export_customers
    proxied: True 
 """)

    dag_proxied_state = DagProxiedState.from_dict(proxied_state_dict)
    assert dag_proxied_state.is_task_proxied("load_raw_customers") is False
    assert dag_proxied_state.is_task_proxied("build_dbt_models") is False
    assert dag_proxied_state.is_task_proxied("export_customers") is True


def test_dag_level_proxied_state_from_yaml() -> None:
    proxied_state_dict = yaml.safe_load("""
proxied: True
""")
    dag_proxied_state = DagProxiedState.from_dict(proxied_state_dict)
    assert dag_proxied_state.proxied is True

    proxied_state_dict = yaml.safe_load("""
proxied: False
""")
    dag_proxied_state = DagProxiedState.from_dict(proxied_state_dict)
    assert dag_proxied_state.proxied is False

    proxied_state_dict = yaml.safe_load("""
proxied: Fish
""")
    with pytest.raises(Exception, match="Expected 'proxied' key to be a boolean or None"):
        DagProxiedState.from_dict(proxied_state_dict)



import pytest
from dagster import AssetKey

from kitchen_sink_tests.integration_tests.conftest import (
    ExpectedMat,
    makefile_dir,
    poll_for_expected_mats,
)


@pytest.fixture(name="dagster_dev_cmd")
def dagster_dev_cmd_fixture() -> list[str]:
    return ["make", "run_dagster_multi_code_locations", "-C", str(makefile_dir())]


def test_multiple_code_locations_materialize(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    """Test that assets can load properly, and that materializations register across multiple code locations."""
    from kitchen_sink.airflow_instance import local_airflow_instance

    af_instance = local_airflow_instance()

    expected_mats_per_dag = {
        "dag_first_code_location": [
            ExpectedMat(AssetKey("dag_first_code_location__asset"), runs_in_dagster=False)
        ],
        "dag_second_code_location": [
            ExpectedMat(AssetKey("dag_first_code_location__asset"), runs_in_dagster=False)
        ],
    }

    poll_for_expected_mats(af_instance, expected_mats_per_dag)



from datetime import datetime
from pathlib import Path

from airflow import DAG
from airflow.operators.python import PythonOperator
from dagster_airlift.in_airflow import proxying_to_dagster
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml
from dagster_airlift.in_airflow.task_proxy_operator import DefaultProxyTaskToDagsterOperator


class DeferredEventsTaskProxyOperator(DefaultProxyTaskToDagsterOperator):
    @property
    def should_defer_asset_events(self) -> bool:
        return True


def print_hello() -> None:
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 0,
}


with DAG(
    "deferred_events_dag",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as dag:
    PythonOperator(task_id="print_task", python_callable=print_hello) >> PythonOperator(
        task_id="downstream_print_task", python_callable=print_hello
    )  # type: ignore


proxying_to_dagster(
    proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
    global_vars=globals(),
    build_from_task_fn=DeferredEventsTaskProxyOperator.build_from_task,
)



# ruff: noqa: T201
import argparse
import os
import subprocess
import time
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path

import dagster._check as check
from dagster import Definitions, RepositoryDefinition, build_sensor_context
from dagster._core.instance import DagsterInstance
from dagster._core.test_utils import environ
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.test.shared_fixtures import stand_up_airflow

from perf_harness.shared.constants import CONSTANTS_FILE, get_perf_output_file
from perf_harness.shared.utils import scaffold_proxied_state

MAKEFILE_DIR = Path(__file__).parent.parent
DAGSTER_HOME = MAKEFILE_DIR / ".dagster_home"


@contextmanager
def modify_constants(num_dags, num_tasks, num_assets) -> Generator[None, None, None]:
    # Read the original content
    with open(CONSTANTS_FILE) as f:
        original_content = f.read()

    # Write new constants
    modified_content = (
        f"NUM_DAGS {num_dags}\nNUM_TASKS {num_tasks}\nNUM_ASSETS_PER_TASK {num_assets}\n"
    )

    # Write the modified content
    with open(CONSTANTS_FILE, "w") as f:
        f.write(modified_content)

    try:
        # Yield control back to the caller
        yield
    finally:
        # Restore the original content
        with open(CONSTANTS_FILE, "w") as f:
            f.write(original_content)


def main() -> None:
    lines = []
    parser = argparse.ArgumentParser(description="Performance scenario testing for airlift")
    parser.add_argument("num_dags", type=int, help="Number of DAGs to generate")
    parser.add_argument("num_tasks", type=int, help="Number of tasks per DAG")
    parser.add_argument("num_assets", type=int, help="Number of assets per task")

    args = parser.parse_args()

    num_dags = check.int_param(args.num_dags, "num_dags")
    num_tasks = check.int_param(args.num_tasks, "num_tasks")
    num_assets = check.int_param(args.num_assets, "num_assets")

    with (
        modify_constants(num_dags, num_tasks, num_assets),
        environ({"DAGSTER_HOME": str(DAGSTER_HOME)}),
    ):
        print("Scaffolding proxied state...")
        scaffold_proxied_state(num_dags=num_dags, num_tasks=num_tasks, proxied_state=True)

        print("Importing airflow defs...")
        from perf_harness.airflow_dags.dags import (
            import_time,
            mark_as_dagster_time,
            total_time as total_load_time,
        )

        lines.append(f"Total airflow dags load time: {total_load_time:.4f} seconds\n")
        lines.append(f"Airflow defs import time: {import_time:.4f} seconds\n")
        lines.append(f"Proxying to dagster time: {mark_as_dagster_time:.4f} seconds\n")

        print("Initializing airflow...")
        # Take in as an argument the number of tasks and the number of dags.
        # Create an ephemeral file where the results will be written, and assign it in shared.constants.py
        # Stand up airflow.
        # Stand up dagster at each stage (peer, observe, migrate), wiping .dagster_home in between.
        # At each stage, time how long initial load takes
        # Then, time how long subsequent load takes
        # Run the test sensor graphql mutation and time how long it takes.
        # Write all results to the ephemeral file.

        airflow_setup_time = time.time()
        subprocess.run(
            ["make", "setup_local_env"], check=True, cwd=MAKEFILE_DIR, stdout=subprocess.DEVNULL
        )
        airflow_setup_completion_time = time.time()
        lines.append(
            f"Airflow setup time: {airflow_setup_completion_time - airflow_setup_time:.4f} seconds\n"
        )

        print("Standing up airflow...")
        start_airflow_standup_time = time.time()
        with stand_up_airflow(
            env=os.environ,
            airflow_cmd=["make", "run_airflow"],
            cwd=MAKEFILE_DIR,
            stdout_channel=subprocess.DEVNULL,
        ):
            module_name_to_defs_and_instance = {}
            finished_airflow_standup_time = time.time()
            lines.append(
                f"Airflow standup time: {finished_airflow_standup_time - start_airflow_standup_time:.4f} seconds\n"
            )
            print("Loading peering defs...")
            peered_defs_import_start_time = time.time()
            from perf_harness.dagster_defs.peer import defs

            peered_defs_import_end_time = time.time()
            peered_defs_import_time = peered_defs_import_end_time - peered_defs_import_start_time
            lines.append(f"Peered defs import time: {peered_defs_import_time:.4f} seconds\n")
            from perf_harness.dagster_defs.peer import airflow_instance

            module_name_to_defs_and_instance["peer"] = (defs, airflow_instance)

            print("Loading observing defs...")
            observe_defs_import_start_time = time.time()
            from perf_harness.dagster_defs.observe import defs

            observe_defs_import_end_time = time.time()
            observe_defs_import_time = observe_defs_import_end_time - observe_defs_import_start_time
            lines.append(f"Observed defs import time: {observe_defs_import_time:.4f} seconds\n")
            from perf_harness.dagster_defs.observe import airflow_instance

            module_name_to_defs_and_instance["observe"] = (defs, airflow_instance)

            print("Loading migrated defs...")
            migrate_defs_import_start_time = time.time()
            from perf_harness.dagster_defs.migrate import defs

            migrate_defs_import_end_time = time.time()
            migrate_defs_import_time = migrate_defs_import_end_time - migrate_defs_import_start_time
            lines.append(f"Migrate defs import time: {migrate_defs_import_time:.4f} seconds\n")
            from perf_harness.dagster_defs.migrate import airflow_instance

            module_name_to_defs_and_instance["migrate"] = (defs, airflow_instance)

            print("Running performance tests across modules...")
            lines.extend(
                run_suite_for_defs(
                    module_name_to_defs_and_instance=module_name_to_defs_and_instance,
                    num_dags=num_dags,
                    instance=DagsterInstance.get(),
                )
            )

        with open(get_perf_output_file(), "w") as f:
            f.writelines(lines)
    print("Performance harness completed.")


def run_suite_for_defs(
    *,
    module_name_to_defs_and_instance: dict[str, tuple[Definitions, AirflowInstance]],
    num_dags: int,
    instance: DagsterInstance,
) -> list[str]:
    lines = []
    module_name_to_repo_def_and_instance: dict[
        str, tuple[RepositoryDefinition, AirflowInstance]
    ] = {}
    for module_name, (defs, af_instance) in module_name_to_defs_and_instance.items():
        defs_initial_load_time = time.time()
        print(f"Loading cacheable assets for {module_name} defs...")
        repo_def = defs.get_repository_def()
        defs_initial_load_completion_time = time.time()
        lines.append(
            f"{module_name} defs initial load time: {defs_initial_load_completion_time - defs_initial_load_time:.4f} seconds\n"
        )
        module_name_to_repo_def_and_instance[module_name] = (repo_def, af_instance)

    print("Running sensor ticks with no runs...")
    for module_name, (repo_def, af_instance) in module_name_to_repo_def_and_instance.items():
        sensor_def = repo_def.get_sensor_def("my_airflow_instance__airflow_dag_status_sensor")
        print(f"Running {module_name} sensor ticks...")
        sensor_tick_no_runs_start_time = time.time()
        sensor_def(build_sensor_context(repository_def=repo_def, instance=instance))
        sensor_tick_no_runs_end_time = time.time()

        lines.append(
            f"{module_name} sensor tick with no runs time: {sensor_tick_no_runs_end_time - sensor_tick_no_runs_start_time:.4f} seconds\n"
        )

    print("Running dag once...")
    run_id = af_instance.trigger_dag("dag_0")  # pyright: ignore[reportPossiblyUnboundVariable]
    af_instance.wait_for_run_completion("dag_0", run_id, timeout=400)  # pyright: ignore[reportPossiblyUnboundVariable]
    print("Dag run completed.")
    print("Running sensor ticks with single run...")
    for module_name, (repo_def, af_instance) in module_name_to_repo_def_and_instance.items():
        sensor_def = repo_def.get_sensor_def("my_airflow_instance__airflow_dag_status_sensor")
        sensor_tick_with_single_run_start_time = time.time()
        sensor_def(build_sensor_context(repository_def=repo_def, instance=instance))
        sensor_tick_with_single_run_end_time = time.time()
        lines.append(
            f"{module_name} sensor tick with single run time: {sensor_tick_with_single_run_end_time - sensor_tick_with_single_run_start_time:.4f} seconds\n"
        )
    print("Deleting run...")
    # Delete that run. Then add a run to every dag.
    af_instance.delete_run("dag_0", run_id)  # pyright: ignore[reportPossiblyUnboundVariable]
    print("Running every dag...")
    newly_added_runs = {}
    for i in range(num_dags):
        newly_added_runs[f"dag_{i}"] = af_instance.trigger_dag(f"dag_{i}")  # pyright: ignore[reportPossiblyUnboundVariable]
    completed_runs = []
    while len(completed_runs) < num_dags:
        for dag_id, run_id in newly_added_runs.items():
            if run_id in completed_runs:
                continue
            run_state = af_instance.get_run_state(dag_id, run_id)  # pyright: ignore[reportPossiblyUnboundVariable]
            if run_state == "success":
                completed_runs.append(run_id)
                continue
            if run_state == "failed":
                raise Exception("A run failed.")
    print("All runs completed.")
    print("Running sensor ticks with all runs...")
    for module_name, (repo_def, af_instance) in module_name_to_repo_def_and_instance.items():
        sensor_def = repo_def.get_sensor_def("my_airflow_instance__airflow_dag_status_sensor")
        sensor_tick_with_all_runs_start_time = time.time()
        sensor_def(build_sensor_context(repository_def=repo_def, instance=instance))
        sensor_tick_with_all_runs_end_time = time.time()
        lines.append(
            f"{module_name} sensor tick with {num_dags} runs time: {sensor_tick_with_all_runs_end_time - sensor_tick_with_all_runs_start_time:.4f} seconds\n"
        )
    # Delete all runs.
    for dag_id, run_id in newly_added_runs.items():
        af_instance.delete_run(dag_id, run_id)  # pyright: ignore[reportPossiblyUnboundVariable]
    print("Deleted all runs.")
    return lines


if __name__ == "__main__":
    main()



from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING

from airflow import DAG
from airflow.operators.python import PythonOperator

if TYPE_CHECKING:
    from airflow.models.operator import BaseOperator


def build_dags_dict_given_structure(structure: dict[str, dict[str, list[str]]]) -> dict[str, DAG]:
    """Given a structure of dags and their tasks, build a dictionary of dags."""
    return_dict = {}
    tasks_per_dag: dict[str, dict[str, BaseOperator]] = defaultdict(dict)
    for dag_id, task_structure in structure.items():
        dag = DAG(
            dag_id,
            default_args={
                "owner": "airflow",
                "depends_on_past": False,
                "start_date": datetime(2023, 1, 1),
                "retries": 1,
            },
            schedule_interval=None,
            is_paused_upon_creation=False,
        )
        for task_id in task_structure.keys():
            tasks_per_dag[dag_id][task_id] = PythonOperator(
                task_id=task_id,
                python_callable=lambda: None,
                dag=dag,
            )
        return_dict[dag_id] = dag

    # Do another pass to build out dependencies
    for dag_id, task_structure in structure.items():
        for task_id, deps in task_structure.items():
            for dep in deps:
                tasks_per_dag[dag_id][task_id].set_upstream(tasks_per_dag[dag_id][dep])
    return return_dict



from dagster_airlift.core import (
    AirflowBasicAuthBackend,
    AirflowInstance,
    build_defs_from_airflow_instance,
)

from perf_harness.dagster_defs.constants import (
    AIRFLOW_BASE_URL,
    AIRFLOW_INSTANCE_NAME,
    PASSWORD,
    USERNAME,
)

airflow_instance = AirflowInstance(
    auth_backend=AirflowBasicAuthBackend(
        webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
    ),
    name=AIRFLOW_INSTANCE_NAME,
)


defs = build_defs_from_airflow_instance(airflow_instance=airflow_instance)



import subprocess
from pathlib import Path

import yaml
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml
from dagster_airlift.test import configured_airflow_home


def path_to_test_proj() -> Path:
    return Path(__file__).parent / "scaffold_test_airflow_home"


def run_scaffold_script() -> None:
    subprocess.run(["dagster-airlift", "proxy", "scaffold"], check=False)


def expected_yaml() -> str:
    return """
tasks:
  - id: downstream_print_task
    proxied: false
  - id: print_task 
    proxied: false 
"""


def test_scaffold_proxied_state() -> None:
    """Test scaffold proxied state under different scenarios."""
    # No AIRFLOW_HOME set.

    # Airflow home set but a proxied_state directory already exists.

    # Airflow home set and no proxied_state directory exists.
    with configured_airflow_home(path_to_test_proj()):
        run_scaffold_script()
        path_to_proxied_state = path_to_test_proj() / "airflow_dags" / "proxied_state"
        assert path_to_proxied_state.exists()
        symbols = list(path_to_proxied_state.iterdir())
        # Check that path contents are two files, one for each dag.
        assert len(list(symbols)) == 2
        assert all([symbol.is_file() for symbol in symbols])
        assert all([symbol.suffix == ".yaml" for symbol in symbols])
        assert {symbol.name for symbol in symbols} == {"print_dag.yaml", "other_print_dag.yaml"}
        print_dag_yaml = path_to_proxied_state / "print_dag.yaml"
        assert yaml.safe_load(print_dag_yaml.read_text()) == yaml.safe_load(expected_yaml())
        other_print_dag_yaml = path_to_proxied_state / "other_print_dag.yaml"
        assert yaml.safe_load(other_print_dag_yaml.read_text()) == yaml.safe_load(expected_yaml())
        proxied_state = load_proxied_state_from_yaml(path_to_proxied_state)
        assert (
            proxied_state.get_task_proxied_state(dag_id="print_dag", task_id="print_task") is False
        )
        assert (
            proxied_state.get_task_proxied_state(
                dag_id="print_dag", task_id="downstream_print_task"
            )
            is False
        )
        assert (
            proxied_state.get_task_proxied_state(dag_id="other_print_dag", task_id="print_task")
            is False
        )
        assert (
            proxied_state.get_task_proxied_state(
                dag_id="other_print_dag", task_id="downstream_print_task"
            )
            is False
        )



from datetime import datetime

from airflow import DAG
from airflow.operators.python import PythonOperator


def print_hello() -> None:
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 0,
}


def make_print_dag(dag_id: str) -> DAG:
    with DAG(
        dag_id,
        default_args=default_args,
        schedule_interval=None,
        is_paused_upon_creation=False,
    ) as dag:
        PythonOperator(task_id="print_task", python_callable=print_hello) >> PythonOperator(
            task_id="downstream_print_task", python_callable=print_hello
        )  # type: ignore
        return dag


print_dag = make_print_dag("print_dag")
other_print_dag = make_print_dag("other_print_dag")



import datetime

import pytest
from dagster_airlift.core import AirflowBasicAuthBackend, AirflowInstance
from dagster_airlift.core.filter import AirflowFilter
from dagster_shared.error import DagsterError

from dagster_airlift_tests.integration_tests.conftest import assert_link_exists


def test_airflow_instance(airflow_instance: None) -> None:
    """Test AirflowInstance APIs against live-running airflow. Ensure that links result in 200s.

    Airflow is loaded with one dag (print_dag) which contains two tasks (print_task, downstream_print_task).
    """
    instance = AirflowInstance(
        auth_backend=AirflowBasicAuthBackend(
            webserver_url="http://localhost:8080", username="admin", password="admin"
        ),
        name="airflow_instance",
    )
    dag_infos = instance.list_dags(retrieval_filter=AirflowFilter())
    assert len(dag_infos) == 1
    assert dag_infos[0].dag_id == "print_dag"
    # Required for source code fetching
    assert "file_token" in dag_infos[0].metadata
    source_code = instance.get_dag_source_code(dag_infos[0].metadata["file_token"])
    assert "print_hello()" in source_code
    # Attempt a nonexistent file token
    with pytest.raises(DagsterError, match=r"Failed to fetch source code."):
        instance.get_dag_source_code("nonexistent")

    task_info = instance.get_task_info(dag_id="print_dag", task_id="print_task")
    assert task_info.dag_id == "print_dag"
    assert task_info.task_id == "print_task"
    assert_link_exists("Dag url from task info object", task_info.dag_url)

    task_infos = instance.get_task_infos(dag_id="print_dag")
    assert len(task_infos) == 2

    task_dict = {task_info.task_id: task_info for task_info in task_infos}
    assert set(task_dict.keys()) == {"print_task", "downstream_print_task"}
    assert task_dict["print_task"].dag_id == "print_dag"
    assert task_dict["print_task"].task_id == "print_task"
    assert task_dict["downstream_print_task"].dag_id == "print_dag"
    assert task_dict["downstream_print_task"].task_id == "downstream_print_task"
    assert task_dict["print_task"].downstream_task_ids == ["downstream_print_task"]

    task_info = instance.get_task_info(dag_id="print_dag", task_id="downstream_print_task")
    assert task_info.dag_id == "print_dag"
    assert task_info.task_id == "downstream_print_task"
    assert_link_exists("Dag url from task info object", task_info.dag_url)

    # Attempt a nonexistent task
    with pytest.raises(
        DagsterError, match=r"Failed to fetch task info for print_dag/nonexistent_task."
    ):
        instance.get_task_info(dag_id="print_dag", task_id="nonexistent_task")

    # Kick off a run of the dag.
    run_id = instance.trigger_dag(dag_id="print_dag")
    instance.wait_for_run_completion(dag_id="print_dag", run_id=run_id)
    run = instance.get_dag_run(dag_id="print_dag", run_id=run_id)

    assert run.run_id == run_id
    assert_link_exists("Dag run", run.url)

    assert run.finished
    assert run.success
    assert isinstance(run.start_date, datetime.datetime)
    assert isinstance(run.end_date, datetime.datetime)
    assert isinstance(run.logical_date, datetime.datetime)

    # Fetch task instance
    task_instance = instance.get_task_instance(
        dag_id="print_dag", task_id="print_task", run_id=run_id
    )
    assert_link_exists("Task instance", task_instance.details_url)
    assert_link_exists("Task logs", task_instance.log_url)

    assert isinstance(task_instance.start_date, datetime.datetime)
    assert isinstance(task_instance.end_date, datetime.datetime)
    assert isinstance(task_instance.note, str)
    assert isinstance(task_instance.logical_date, datetime.datetime)
    assert run.logical_date == task_instance.logical_date



from datetime import datetime
from pathlib import Path

from airflow import DAG
from airflow.operators.python import PythonOperator
from dagster_airlift.in_airflow import proxying_to_dagster
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml


def print_hello() -> None:
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 0,
}

with DAG(
    "dag_first_code_location",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as first_dag:
    PythonOperator(task_id="task", python_callable=print_hello)

with DAG(
    "dag_second_code_location",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as second_dag:
    PythonOperator(task_id="task", python_callable=print_hello)


proxying_to_dagster(
    proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
    global_vars=globals(),
)



from collections.abc import Sequence
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest
from dagster import (
    AssetCheckKey,
    AssetKey,
    AssetSpec,
    DagsterInstance,
    Definitions,
    SensorResult,
    asset,
    asset_check,
    build_sensor_context,
)
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.materialize import materialize
from dagster._core.definitions.metadata.metadata_value import TimestampMetadataValue
from dagster._core.definitions.partitions.definition import DailyPartitionsDefinition
from dagster._core.definitions.sensor_definition import SensorEvaluationContext
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.execution.execute_in_process_result import ExecuteInProcessResult
from dagster._core.test_utils import freeze_time
from dagster._serdes import deserialize_value
from dagster._time import get_current_datetime
from dagster_airlift.constants import (
    DAG_ID_TAG_KEY,
    DAG_MAPPING_METADATA_KEY,
    DAG_RUN_ID_TAG_KEY,
    EFFECTIVE_TIMESTAMP_METADATA_KEY,
    TASK_ID_TAG_KEY,
)
from dagster_airlift.core import build_defs_from_airflow_instance
from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData
from dagster_airlift.core.sensor.sensor_builder import (
    AirflowPollingSensorCursor,
    AirliftSensorEventTransformerError,
)
from dagster_airlift.core.top_level_dag_def_api import assets_with_task_mappings
from dagster_airlift.test import make_dag_run, make_instance

from dagster_airlift_tests.unit_tests.conftest import (
    assert_expected_key_order,
    build_and_invoke_sensor,
    fully_loaded_repo_from_airflow_asset_graph,
)


def make_dag_key(dag_id: str) -> AssetKey:
    return AssetKey(["test_instance", "dag", dag_id])


def make_dag_key_str(dag_id: str) -> str:
    return make_dag_key(dag_id).to_user_string()


def test_dag_and_task_metadata(init_load_context: None, instance: DagsterInstance) -> None:
    """Test the metadata produced by a sensor for a single dag and task."""
    freeze_datetime = datetime(2021, 1, 1)

    with freeze_time(freeze_datetime):
        result, _ = build_and_invoke_sensor(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            instance=instance,
        )
        assert len(result.asset_events) == 2
        assert_expected_key_order(result.asset_events, ["a", make_dag_key_str("dag")])
        dag_mat = result.asset_events[1]
        expected_dag_metadata_keys = {
            "Airflow Run ID",
            "Start Date",
            "End Date",
            "Run Metadata (raw)",
            "Creation Timestamp",
            "Run Details",
            "Airflow Config",
            "Run Type",
            "dagster-airlift/effective-timestamp",
            "dagster-airlift/airflow-run-id",
        }
        assert set(dag_mat.metadata.keys()) == expected_dag_metadata_keys
        task_mat = result.asset_events[0]
        expected_task_metadata_keys = {
            "Airflow Run ID",
            "Start Date",
            "End Date",
            "Run Metadata (raw)",
            "Creation Timestamp",
            "Run Details",
            "Task Logs",
            "Airflow Config",
            "Run Type",
            "dagster-airlift/effective-timestamp",
            "dagster-airlift/airflow-run-id",
            "dagster-airlift/airflow-task-instance-logical-date",
        }
        assert set(task_mat.metadata.keys()) == expected_task_metadata_keys

        assert task_mat.metadata["Airflow Run ID"].value == "run-dag"
        assert (
            task_mat.metadata["Start Date"].value
            == (freeze_datetime - timedelta(minutes=10)).timestamp()
        )
        assert (
            task_mat.metadata["End Date"].value
            == (freeze_datetime - timedelta(seconds=1)).timestamp()
        )
        assert task_mat.metadata["Creation Timestamp"].value == freeze_datetime.timestamp()
        assert (
            task_mat.metadata["Run Details"].value
            == "[View Run](http://dummy.domain/dags/dag/grid?dag_run_id=run-dag&task_id=task)"
        )
        assert (
            task_mat.metadata["Task Logs"].value
            == "[View Logs](http://dummy.domain/dags/dag/grid?dag_run_id=run-dag&task_id=task&tab=logs)"
        )
        assert task_mat.metadata["Airflow Config"].value == {}
        assert task_mat.metadata["Run Type"].value == "manual"

        assert dag_mat.metadata["Airflow Run ID"].value == "run-dag"
        assert (
            dag_mat.metadata["Start Date"].value
            == (freeze_datetime - timedelta(minutes=10)).timestamp()
        )
        assert dag_mat.metadata["End Date"].value == freeze_datetime.timestamp()
        assert dag_mat.metadata["Creation Timestamp"].value == freeze_datetime.timestamp()
        assert (
            dag_mat.metadata["Run Details"].value
            == "[View Run](http://dummy.domain/dags/dag/grid?dag_run_id=run-dag&tab=details)"
        )
        assert dag_mat.metadata["Airflow Config"].value == {}
        assert dag_mat.metadata["Run Type"].value == "manual"


def test_interleaved_executions(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that the when task / dag completion is interleaved the correct ordering is preserved."""
    # Asset graph structure:
    #   a -> b where a and b are each in their own airflow tasks.
    #   c -> d where c and d are each in their own airflow tasks, in a different dag.
    freeze_datetime = datetime(2021, 1, 1)
    with freeze_time(freeze_datetime):
        result, context = build_and_invoke_sensor(
            assets_per_task={
                "dag1": {"task1": [("a", [])], "task2": [("b", ["a"])]},
                "dag2": {"task1": [("c", [])], "task2": [("d", ["c"])]},
            },
            instance=instance,
        )
        # We expect one asset materialization per asset.
        assert len(result.asset_events) == 6
        assert all(isinstance(event, AssetMaterialization) for event in result.asset_events)

        mats_order = [mat.asset_key.to_user_string() for mat in result.asset_events]
        # a should be before b
        assert mats_order.index("a") < mats_order.index("b")
        # c should be before d
        assert mats_order.index("c") < mats_order.index("d")
        # dag1 and dag2 should be after all task-mapped assets
        assert mats_order.index(make_dag_key_str("dag1")) >= 4
        assert mats_order.index(make_dag_key_str("dag2")) >= 4
        assert context.cursor
        cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert cursor.end_date_gte == freeze_datetime.timestamp()
        assert cursor.end_date_lte is None
        assert cursor.dag_query_offset == 0


def test_dependencies_within_tasks(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that a complex asset graph structure can be ingested in correct order from the sensor.
    Where a, b, and c are part of task 1, and d, e, and f are part of task 2.
    """
    # Asset graph structure:
    #   a
    #  / \
    # b   c
    #  \ /
    #   d
    #  / \
    # e   f
    freeze_datetime = datetime(2021, 1, 1)
    with freeze_time(freeze_datetime):
        result, context = build_and_invoke_sensor(
            assets_per_task={
                "dag": {
                    "task1": [("a", []), ("b", ["a"]), ("c", ["a"])],
                    "task2": [("d", ["b", "c"]), ("e", ["d"]), ("f", ["d"])],
                },
            },
            instance=instance,
        )
        assert len(result.asset_events) == 7
        assert_expected_key_order(
            result.asset_events, ["a", "b", "c", "d", "e", "f", make_dag_key_str("dag")]
        )
        assert context.cursor
        cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert cursor.end_date_gte == freeze_datetime.timestamp()
        assert cursor.end_date_lte is None
        assert cursor.dag_query_offset == 0


def test_outside_of_dag_dependency(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that if an asset has a transitive dependency on another asset within the same task, ordering is respected."""
    # a -> b -> c where a and c are in the same task, and b is not in any dag.
    freeze_datetime = datetime(2021, 1, 1)
    with freeze_time(freeze_datetime):
        result, context = build_and_invoke_sensor(
            assets_per_task={
                "dag": {"task": [("a", []), ("c", ["b"])]},
            },
            additional_defs=Definitions(assets=[AssetSpec(key="b", deps=["a"])]),
            instance=instance,
        )
        assert len(result.asset_events) == 3
        assert all(isinstance(event, AssetMaterialization) for event in result.asset_events)
        assert_expected_key_order(result.asset_events, ["a", "c", make_dag_key_str("dag")])
        assert context.cursor
        cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert cursor.end_date_gte == freeze_datetime.timestamp()
        assert cursor.end_date_lte is None
        assert cursor.dag_query_offset == 0


def test_request_asset_checks(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that when a new dag or task run is detected, a new check run is requested for all checks which may target that dag/task."""
    freeze_datetime = datetime(2021, 1, 1)

    dag_asset_key = make_dag_key("dag")

    @asset_check(asset="a")  # pyright: ignore[reportArgumentType]
    def check_task_asset():
        pass

    @asset_check(asset=dag_asset_key)  # pyright: ignore[reportArgumentType]
    def check_dag_asset():
        pass

    @asset_check(asset="c")  # pyright: ignore[reportArgumentType]
    def check_unrelated_asset():
        pass

    with freeze_time(freeze_datetime):
        result, context = build_and_invoke_sensor(
            assets_per_task={
                "dag": {"task": [("a", []), ("b", ["a"])]},
            },
            additional_defs=Definitions(
                asset_checks=[check_task_asset, check_dag_asset, check_unrelated_asset],
                assets=[AssetSpec(key="c")],
            ),
            instance=instance,
        )

        assert len(result.asset_events) == 3
        assert result.run_requests
        assert len(result.run_requests) == 1
        run_request = result.run_requests[0]
        assert run_request.asset_check_keys
        assert set(run_request.asset_check_keys) == {
            AssetCheckKey(name="check_task_asset", asset_key=AssetKey(["a"])),
            AssetCheckKey(name="check_dag_asset", asset_key=dag_asset_key),
        }
        assert context.cursor
        cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert cursor.end_date_gte == freeze_datetime.timestamp()
        assert cursor.end_date_lte is None
        assert cursor.dag_query_offset == 0


_CALLCOUNT = [0]


def _create_datetime_mocker(iter_times: list[datetime]):
    def _mock_get_current_datetime() -> datetime:
        the_time = iter_times[_CALLCOUNT[0]]
        _CALLCOUNT[0] += 1
        return the_time

    return _mock_get_current_datetime


def test_cursor(init_load_context: None, instance: DagsterInstance) -> None:
    """Test expected cursor behavior for sensor."""
    asset_and_dag_structure = {
        "dag1": {"task1": [("a", [])]},
        "dag2": {"task1": [("b", [])]},
    }

    with freeze_time(datetime(2021, 1, 1, tzinfo=timezone.utc)):
        # First, run through a full successful iteration of the sensor.
        # Expect time to move forward, and polled dag id to be None, since we completed iteration of all dags.
        # Then, run through a partial iteration of the sensor. We mock get_current_datetime to return a time
        # after timeout passes iteration start after the first call, meaning we should pause iteration.
        repo_def = fully_loaded_repo_from_airflow_asset_graph(asset_and_dag_structure)
        sensor = next(iter(repo_def.sensor_defs))
        context = build_sensor_context(repository_def=repo_def, instance=instance)
        result = sensor(context)
        assert isinstance(result, SensorResult)
        assert context.cursor
        new_cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert new_cursor.end_date_gte == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.end_date_lte is None
        assert new_cursor.dag_query_offset == 0

    # Now, we expect that we will not have completed iteration before we need to pause evaluation.
    datetimes = [
        datetime(2021, 2, 1, tzinfo=timezone.utc),  # set initial time
        datetime(2021, 2, 1, 0, 0, 30, tzinfo=timezone.utc),  # initial iteration time
        datetime(
            2022, 2, 2, tzinfo=timezone.utc
        ),  # second iteration time, at which iteration should be paused
    ]
    with mock.patch(
        "dagster._time._mockable_get_current_datetime", wraps=_create_datetime_mocker(datetimes)
    ):
        result = sensor(context)
        assert isinstance(result, SensorResult)
        new_cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        # We didn't advance to the next effective timestamp, since we didn't complete iteration
        assert new_cursor.end_date_gte == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
        # We have not yet moved forward
        assert new_cursor.end_date_lte == datetime(2021, 2, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.dag_query_offset == 1

        _CALLCOUNT[0] = 0
        # We weren't able to complete iteration, so we should pause iteration again
        result = sensor(context)
        assert isinstance(result, SensorResult)
        new_cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert new_cursor.end_date_gte == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.end_date_lte == datetime(2021, 2, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.dag_query_offset == 2

        _CALLCOUNT[0] = 0
        # Now it should finish iteration.
        result = sensor(context)
        assert isinstance(result, SensorResult)
        new_cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert new_cursor.end_date_gte == datetime(2021, 2, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.end_date_lte is None
        assert new_cursor.dag_query_offset == 0


def test_legacy_cursor(init_load_context: None, instance: DagsterInstance) -> None:
    """Test the case where a legacy/uninterpretable cursor is provided to the sensor execution."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)
    with freeze_time(freeze_datetime):
        repo_def = fully_loaded_repo_from_airflow_asset_graph({"dag": {"task": [("a", [])]}})
        sensor = next(iter(repo_def.sensor_defs))
        context = build_sensor_context(
            repository_def=repo_def, cursor=str(freeze_datetime.timestamp()), instance=instance
        )
        result = sensor(context)
        assert isinstance(result, SensorResult)
        assert context.cursor
        new_cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert new_cursor.end_date_gte == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.end_date_lte is None
        assert new_cursor.dag_query_offset == 0


def test_no_runs(init_load_context: None, instance: DagsterInstance) -> None:
    """Test the case with no runs."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)
    with freeze_time(freeze_datetime):
        repo_def = fully_loaded_repo_from_airflow_asset_graph(
            {"dag": {"task": [("a", [])]}},
            create_runs=False,
        )
        sensor = next(iter(repo_def.sensor_defs))
        context = build_sensor_context(repository_def=repo_def, instance=instance)
        result = sensor(context)
        assert isinstance(result, SensorResult)
        assert context.cursor
        new_cursor = deserialize_value(context.cursor, AirflowPollingSensorCursor)
        assert new_cursor.end_date_gte == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
        assert new_cursor.end_date_lte is None
        assert new_cursor.dag_query_offset == 0
        assert not result.asset_events
        assert not result.run_requests


def simulate_materialize_from_proxy_operator(
    *,
    instance: DagsterInstance,
    assets_def: AssetsDefinition,
    dag_run_id: str,
    dag_id: str,
    task_id: str | None,
) -> ExecuteInProcessResult:
    # TODO consolidate with code in proxy operator
    tags = {
        DAG_ID_TAG_KEY: dag_id,
        DAG_RUN_ID_TAG_KEY: dag_run_id,
    }
    if task_id:
        tags[TASK_ID_TAG_KEY] = task_id
    return materialize(assets=[assets_def], instance=instance, tags=tags)


def test_pluggable_transformation(init_load_context: None, instance: DagsterInstance) -> None:
    """Test the case where a custom transformation is provided to the sensor."""

    def pluggable_event_transformer(
        context: SensorEvaluationContext,
        airflow_data: AirflowDefinitionsData,
        events: Sequence[AssetMaterialization],
    ) -> Sequence[AssetMaterialization]:
        assert isinstance(context, SensorEvaluationContext)
        assert isinstance(airflow_data, AirflowDefinitionsData)
        # Change the timestamp, which should also change the order. We expect this to be respected by the sensor.
        new_events = []
        for event in events:
            if AssetKey(["a"]) == event.asset_key:
                new_events.append(
                    event._replace(
                        metadata={
                            "test": "test",
                            EFFECTIVE_TIMESTAMP_METADATA_KEY: TimestampMetadataValue(1.0),
                        }
                    )
                )
            elif AssetKey.from_user_string(make_dag_key_str("dag")) == event.asset_key:
                new_events.append(
                    event._replace(
                        metadata={
                            "test": "test",
                            EFFECTIVE_TIMESTAMP_METADATA_KEY: TimestampMetadataValue(0.0),
                        }
                    )
                )
        return new_events

    result, _context = build_and_invoke_sensor(
        assets_per_task={
            "dag": {"task": [("a", [])]},
        },
        instance=instance,
        event_transformer_fn=pluggable_event_transformer,
    )
    assert len(result.asset_events) == 2
    assert_expected_key_order(result.asset_events, [make_dag_key_str("dag"), "a"])
    for event in result.asset_events:
        assert set(event.metadata.keys()) == {"test", EFFECTIVE_TIMESTAMP_METADATA_KEY}


def test_user_code_error_pluggable_transformation(
    init_load_context: None, instance: DagsterInstance
) -> None:
    """Test the case where a custom transformation is provided to the sensor, and the user code raises an error."""

    def pluggable_event_transformer(
        context: SensorEvaluationContext,
        airflow_data: AirflowDefinitionsData,
        events: Sequence[AssetMaterialization],
    ) -> Sequence[AssetMaterialization]:
        raise ValueError("User code error")

    with pytest.raises(AirliftSensorEventTransformerError):
        build_and_invoke_sensor(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            instance=instance,
            event_transformer_fn=pluggable_event_transformer,
        )


def test_missing_effective_timestamp_pluggable_impl(
    init_load_context: None, instance: DagsterInstance
) -> None:
    """Test the case where a custom transformation is provided to the sensor, and the user doesn't include effective timestamp metadata."""

    def missing_effective_timestamp(
        context: SensorEvaluationContext,
        airflow_data: AirflowDefinitionsData,
        events: Sequence[AssetMaterialization],
    ) -> Sequence[AssetMaterialization]:
        return [event._replace(metadata={}) for event in events]

    with pytest.raises(DagsterInvariantViolationError):
        build_and_invoke_sensor(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            instance=instance,
            event_transformer_fn=missing_effective_timestamp,
        )


def test_nonsense_result_pluggable_impl(init_load_context: None, instance: DagsterInstance) -> None:
    """Test the case where a nonsense result is returned from the custom transformation."""

    def nonsense_result(
        context: SensorEvaluationContext,
        airflow_data: AirflowDefinitionsData,
        events: Sequence[AssetMaterialization],
    ) -> Sequence[AssetMaterialization]:
        return [1, 2, 3]  # type: ignore # intentionally wrong

    with pytest.raises(DagsterInvariantViolationError):
        build_and_invoke_sensor(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            instance=instance,
            event_transformer_fn=nonsense_result,
        )


def test_dag_level_override_materializations(
    init_load_context: None, instance: DagsterInstance
) -> None:
    """Test that when a dag level override is provided, the sensor adds the expected asset materializations, in the correct order."""
    freeze_datetime = datetime(2021, 1, 1)

    with freeze_time(freeze_datetime):
        result, _context = build_and_invoke_sensor(
            assets_per_task={},
            dag_level_asset_overrides={"dag": ["a", "b"]},
            instance=instance,
        )
        assert len(result.asset_events) == 3  # 2 for the dag level overrides, 1 for the dag itself
        assert set([event.asset_key.to_user_string() for event in result.asset_events]) == {
            "a",
            "b",
            make_dag_key_str("dag"),
        }
        key_order = [event.asset_key.to_user_string() for event in result.asset_events]
        assert key_order.index("a") < key_order.index(make_dag_key_str("dag"))
        assert key_order.index("b") < key_order.index(make_dag_key_str("dag"))


def test_dag_level_override_existing_runs(
    init_load_context: None, instance: DagsterInstance
) -> None:
    """Test that when there is a run mapping to the same dag, the sensor doesn't add synthetic materializations to mapped assets (only the dag-level peered asset)."""
    freeze_datetime = datetime(2021, 1, 1)

    with freeze_time(freeze_datetime):
        dag_id = "dag"
        dag_run_id = f"run-{dag_id}"
        dag_runs = [
            make_dag_run(
                dag_id=dag_id,
                run_id=f"run-{dag_id}",
                start_date=get_current_datetime() - timedelta(minutes=10),
                end_date=get_current_datetime(),
            )
        ]

        airflow_instance = make_instance(
            dag_and_task_structure={dag_id: ["task"]},
            dag_runs=dag_runs,
            instance_name="test_instance",
        )

        @asset(metadata={DAG_MAPPING_METADATA_KEY: [{"dag_id": dag_id}]})
        def explicit_asset1() -> None:
            pass

        @asset(metadata={DAG_MAPPING_METADATA_KEY: [{"dag_id": dag_id}]})
        def explicit_asset2() -> None:
            pass

        # Simulate a proxied execution for the assets, meaning we don't need to synthesize materializations
        for assets_def in [explicit_asset1, explicit_asset2]:
            assert simulate_materialize_from_proxy_operator(
                instance=instance,
                assets_def=assets_def,
                dag_id=dag_id,
                task_id=None,
                dag_run_id=dag_run_id,
            ).success

        defs = build_defs_from_airflow_instance(
            airflow_instance=airflow_instance,
            defs=Definitions(
                assets=[explicit_asset1, explicit_asset2],
            ),
        )
        assert defs.sensors
        sensor = next(iter(defs.sensors))
        sensor_context = build_sensor_context(
            instance=instance, repository_def=defs.get_repository_def()
        )
        result = sensor(sensor_context)
        assert isinstance(result, SensorResult)
        assert len(result.asset_events) == 1
        assert result.asset_events[0].asset_key.to_user_string() == make_dag_key_str(dag_id)


def test_default_time_partitioned_asset(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that a task instance for a time-partitioned asset is correctly ingested."""
    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            dag_and_task_structure={
                "dag": ["task"],
            },
            dag_runs=[
                make_dag_run(
                    dag_id="dag",
                    run_id="run-dag",
                    start_date=datetime(2021, 1, 2, 5, tzinfo=timezone.utc),
                    end_date=datetime(2021, 1, 2, 6, tzinfo=timezone.utc),
                    logical_date=datetime(2021, 1, 1, tzinfo=timezone.utc),
                )
            ],
        ),
        defs=Definitions(
            assets=assets_with_task_mappings(
                dag_id="dag",
                task_mappings={
                    "task": [
                        AssetSpec(
                            key="a",
                            partitions_def=DailyPartitionsDefinition(
                                start_date=datetime(2021, 1, 1, tzinfo=timezone.utc)
                            ),
                        )
                    ],
                },
            ),
        ),
    )
    assert defs.sensors
    sensor = next(iter(defs.sensors))
    with freeze_time(datetime(2021, 1, 2, 6, tzinfo=timezone.utc)):
        sensor_context = build_sensor_context(
            repository_def=defs.get_repository_def(), instance=instance
        )
        result = sensor(sensor_context)
        assert isinstance(result, SensorResult)
        assert len(result.asset_events) == 2
        assert_expected_key_order(result.asset_events, ["a", "test_instance/dag/dag"])
        a_asset_mat = result.asset_events[0]
        assert isinstance(a_asset_mat, AssetMaterialization)
        # We expect the partition to match the logical date.
        assert a_asset_mat.partition == "2021-01-01"


def test_before_start_of_partitioned_asset(
    init_load_context: None, instance: DagsterInstance
) -> None:
    """We expect to throw an error if there is no matching partition after the start date."""
    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            dag_and_task_structure={
                "dag": ["task"],
            },
            dag_runs=[
                make_dag_run(
                    dag_id="dag",
                    run_id="run-dag",
                    start_date=datetime(2021, 1, 1, 5, tzinfo=timezone.utc),
                    end_date=datetime(2021, 1, 1, 6, tzinfo=timezone.utc),
                    logical_date=datetime(2021, 1, 1, tzinfo=timezone.utc),
                )
            ],
        ),
        defs=Definitions(
            assets=assets_with_task_mappings(
                dag_id="dag",
                task_mappings={
                    "task": [
                        AssetSpec(
                            key="a",
                            partitions_def=DailyPartitionsDefinition(
                                # Partitions definition starts after the logical date.
                                start_date=datetime(2021, 1, 2, tzinfo=timezone.utc)
                            ),
                        )
                    ],
                },
            ),
        ),
    )
    assert defs.sensors
    sensor = next(iter(defs.sensors))
    with freeze_time(datetime(2021, 1, 1, 6, tzinfo=timezone.utc)):
        sensor_context = build_sensor_context(
            repository_def=defs.get_repository_def(), instance=instance
        )
        with pytest.raises(AirliftSensorEventTransformerError):
            sensor(sensor_context)


def test_logical_date_mismatch(init_load_context: None, instance: DagsterInstance) -> None:
    """Test a logical date which does not align with the partition definition due to date mismatch."""
    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            dag_and_task_structure={
                "dag": ["task"],
            },
            dag_runs=[
                make_dag_run(
                    dag_id="dag",
                    run_id="run-dag",
                    start_date=datetime(2021, 1, 1, 5, tzinfo=timezone.utc),
                    end_date=datetime(2021, 1, 1, 6, tzinfo=timezone.utc),
                    # Logical date isn't aligned with midnight.
                    logical_date=datetime(2021, 1, 1, 3, tzinfo=timezone.utc),
                )
            ],
        ),
        defs=Definitions(
            assets=assets_with_task_mappings(
                dag_id="dag",
                task_mappings={
                    "task": [
                        AssetSpec(
                            key="a",
                            partitions_def=DailyPartitionsDefinition(
                                start_date=datetime(2021, 1, 1, tzinfo=timezone.utc)
                            ),
                        )
                    ],
                },
            ),
        ),
    )
    assert defs.sensors
    sensor = next(iter(defs.sensors))
    with freeze_time(datetime(2021, 1, 1, 6, tzinfo=timezone.utc)):
        sensor_context = build_sensor_context(
            repository_def=defs.get_repository_def(), instance=instance
        )
        with pytest.raises(AirliftSensorEventTransformerError):
            sensor(sensor_context)


def test_partition_offset_mismatch(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that partition offsets are respected when determining the partition corresponding to a logical date."""
    airflow_instance = make_instance(
        dag_and_task_structure={
            "dag": ["task"],
        },
        dag_runs=[
            make_dag_run(
                dag_id="dag",
                run_id="run-dag",
                start_date=datetime(2021, 1, 1, 5, tzinfo=timezone.utc),
                end_date=datetime(2021, 1, 1, 6, tzinfo=timezone.utc),
                # Logical date is 3 AM.
                logical_date=datetime(2021, 1, 1, 3, tzinfo=timezone.utc),
            )
        ],
    )

    defs = build_defs_from_airflow_instance(
        airflow_instance=airflow_instance,
        defs=Definitions(
            assets=assets_with_task_mappings(
                dag_id="dag",
                task_mappings={
                    "task": [
                        AssetSpec(
                            key="a",
                            partitions_def=DailyPartitionsDefinition(
                                start_date=datetime(2021, 1, 1, tzinfo=timezone.utc),
                                hour_offset=6,
                            ),
                        )
                    ],
                },
            ),
        ),
    )
    # Due to differing hours offset, we expect an error to throw.
    assert defs.sensors
    sensor = next(iter(defs.sensors))
    with freeze_time(datetime(2021, 1, 1, 6, tzinfo=timezone.utc)):
        sensor_context = build_sensor_context(
            repository_def=defs.get_repository_def(), instance=instance
        )
        with pytest.raises(AirliftSensorEventTransformerError):
            sensor(sensor_context)

    # now, align the offset and expect success.
    defs = build_defs_from_airflow_instance(
        airflow_instance=airflow_instance,
        defs=Definitions(
            assets=assets_with_task_mappings(
                dag_id="dag",
                task_mappings={
                    "task": [
                        AssetSpec(
                            key="a",
                            partitions_def=DailyPartitionsDefinition(
                                start_date=datetime(2021, 1, 1, tzinfo=timezone.utc),
                                hour_offset=3,
                            ),
                        )
                    ],
                },
            ),
        ),
    )
    assert defs.sensors
    sensor = next(iter(defs.sensors))
    with freeze_time(datetime(2021, 1, 1, 6, tzinfo=timezone.utc)):
        sensor_context = build_sensor_context(
            repository_def=defs.get_repository_def(), instance=instance
        )
        result = sensor(sensor_context)
        assert isinstance(result, SensorResult)
        assert len(result.asset_events) == 2
        assert_expected_key_order(result.asset_events, ["a", "test_instance/dag/dag"])
        a_asset_mat = result.asset_events[0]
        assert isinstance(a_asset_mat, AssetMaterialization)
        # We expect the partition to match the logical date.
        assert a_asset_mat.partition == "2021-01-01"



from dagster import Definitions, asset
from dagster_airlift.core.utils import metadata_for_task_mapping


@asset(metadata=metadata_for_task_mapping(dag_id="the_dag", task_id="some_task"))
def the_dag__some_task():
    return "asset_value"


@asset
def unrelated():
    return "unrelated_value"


@asset(metadata=metadata_for_task_mapping(dag_id="the_dag", task_id="other_task"))
def the_dag__other_task():
    return "other_task_value"


defs = Definitions(assets=[the_dag__other_task, the_dag__some_task, unrelated])



from collections.abc import Iterable, Sequence

from dagster import (
    AssetMaterialization,
    AssetObservation,
    AssetSpec,
    Definitions,
    SensorEvaluationContext,
)
from dagster_airlift.core import (
    AirflowDefinitionsData,
    AssetEvent,
    build_defs_from_airflow_instance,
)
from dagster_airlift.core.top_level_dag_def_api import assets_with_task_mappings

from kitchen_sink.airflow_instance import local_airflow_instance


def observations_from_materializations(
    context: SensorEvaluationContext,
    airflow_data: AirflowDefinitionsData,
    materializations: Sequence[AssetMaterialization],
) -> Iterable[AssetEvent]:
    """Construct AssetObservations from AssetMaterializations."""
    for materialization in materializations:
        yield AssetObservation(
            asset_key=materialization.asset_key,
            description=materialization.description,
            metadata=materialization.metadata,
            tags=materialization.tags,
        )


def build_mapped_defs() -> Definitions:
    return build_defs_from_airflow_instance(
        airflow_instance=local_airflow_instance(),
        defs=Definitions(
            assets=assets_with_task_mappings(
                dag_id="simple_unproxied_dag",
                task_mappings={
                    "print_task": [AssetSpec("my_asset")],
                    "downstream_print_task": [AssetSpec("my_downstream_asset", deps=["my_asset"])],
                },
            ),
        ),
        event_transformer_fn=observations_from_materializations,
        sensor_minimum_interval_seconds=1,
    )


defs = build_mapped_defs()



from dagster import AssetKey, AssetSpec, Definitions, multi_asset
from dagster._core.definitions.asset_key import CoercibleToAssetKey
from dagster_airlift.constants import DAG_MAPPING_METADATA_KEY, TASK_MAPPING_METADATA_KEY
from dagster_airlift.core import assets_with_dag_mappings, assets_with_task_mappings


def from_specs(*specs: AssetSpec) -> Definitions:
    return Definitions(assets=specs)


def asset_spec(defs: Definitions, key: CoercibleToAssetKey) -> AssetSpec:
    ak = AssetKey.from_coercible(key)
    return defs.resolve_assets_def(ak).get_asset_spec(ak)


def has_single_task_handle(spec: AssetSpec, dag_id: str, task_id: str) -> bool:
    assert len(spec.metadata[TASK_MAPPING_METADATA_KEY]) == 1
    task_handle_dict = next(iter(spec.metadata[TASK_MAPPING_METADATA_KEY]))
    return task_handle_dict["dag_id"] == dag_id and task_handle_dict["task_id"] == task_id


def has_single_dag_handle(spec: AssetSpec, dag_id: str) -> bool:
    assert len(spec.metadata[DAG_MAPPING_METADATA_KEY]) == 1
    mapping = next(iter(spec.metadata[DAG_MAPPING_METADATA_KEY]))
    return mapping == {"dag_id": dag_id}


# Replaced with task_mappings equivalent
def test_task_mappings_spec_basic() -> None:
    defs = Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_one",
            task_mappings={
                "task_one": [AssetSpec(key="asset_one")],
            },
        )
    )
    assert has_single_task_handle(asset_spec(defs, "asset_one"), "dag_one", "task_one")


def test_task_mappings_spec() -> None:
    defs = Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_one",
            task_mappings={
                "task_one": [AssetSpec(key="asset_one")],
            },
        )
    )
    assert has_single_task_handle(asset_spec(defs, "asset_one"), "dag_one", "task_one")


# Replaced with task_mappings equivalent
def test_task_mappings_multi_tasks_multi_specs_alternate() -> None:
    defs = Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_one",
            task_mappings={
                "task_one": [AssetSpec(key="asset_one")],
                "task_two": [AssetSpec(key="asset_two"), AssetSpec(key="asset_three")],
            },
        )
    )
    assert has_single_task_handle(asset_spec(defs, "asset_one"), "dag_one", "task_one")
    assert has_single_task_handle(asset_spec(defs, "asset_two"), "dag_one", "task_two")
    assert has_single_task_handle(asset_spec(defs, "asset_three"), "dag_one", "task_two")


def test_task_mappings_multi_tasks_multi_specs() -> None:
    defs = Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_one",
            task_mappings={
                "task_one": [AssetSpec(key="asset_one")],
                "task_two": [AssetSpec(key="asset_two"), AssetSpec(key="asset_three")],
            },
        )
    )
    assert has_single_task_handle(asset_spec(defs, "asset_one"), "dag_one", "task_one")
    assert has_single_task_handle(asset_spec(defs, "asset_two"), "dag_one", "task_two")
    assert has_single_task_handle(asset_spec(defs, "asset_three"), "dag_one", "task_two")


# Replaced with task_mappings equivalent
def test_task_mappings_assets_def_alternate() -> None:
    @multi_asset(specs=[AssetSpec(key="asset_one")])
    def an_asset() -> None: ...

    defs = Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_one",
            task_mappings={"task_one": [an_asset]},
        )
    )
    assert has_single_task_handle(asset_spec(defs, "asset_one"), "dag_one", "task_one")


def test_task_mappings_assets_def() -> None:
    @multi_asset(specs=[AssetSpec(key="asset_one")])
    def an_asset() -> None: ...

    defs = Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_one",
            task_mappings={"task_one": [an_asset]},
        )
    )
    assert has_single_task_handle(asset_spec(defs, "asset_one"), "dag_one", "task_one")


def test_dag_mappings_assets_def() -> None:
    @multi_asset(specs=[AssetSpec(key="asset_one")])
    def an_asset() -> None: ...

    defs = Definitions(
        assets=assets_with_dag_mappings(
            {"dag_one": [an_asset]},
        )
    )
    assert has_single_dag_handle(asset_spec(defs, "asset_one"), "dag_one")



# Airflow instance running at localhost:8080


AIRFLOW_BASE_URL = "http://localhost:8080"
AIRFLOW_INSTANCE_NAME = "my_airflow_instance"

# Authentication credentials (lol)
USERNAME = "admin"
PASSWORD = "admin"



from dagster import AssetSpec, Definitions
from dagster_airlift.core import assets_with_task_mappings, build_defs_from_airflow_instance

from kitchen_sink.airflow_instance import local_airflow_instance

defs = build_defs_from_airflow_instance(
    airflow_instance=local_airflow_instance(),
    defs=Definitions(
        assets=assets_with_task_mappings(
            dag_id="dag_first_code_location",
            task_mappings={
                "task": [AssetSpec(key="dag_first_code_location__asset")],
            },
        ),
    ),
    dag_selector_fn=lambda dag_info: dag_info.dag_id == "dag_first_code_location",
    sensor_minimum_interval_seconds=1,
)



from collections.abc import Mapping
from pathlib import Path
from typing import Any

from airflow import DAG
from airflow.operators.python import PythonOperator
from dagster._time import get_current_datetime_midnight
from dagster_airlift.in_airflow import DefaultProxyTaskToDagsterOperator, proxying_to_dagster
from dagster_airlift.in_airflow.dagster_run_utils import (
    MAX_RETRIES_TAG,
    RETRY_ON_ASSET_OR_OP_FAILURE_TAG,
)
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml


def print_hello() -> None:
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "retries": 0,
}


# Normally this isn't needed, but we're trying to get away with not using a multi-process-safe run storage
# to test behavior here.
class SetDagsterRetryInfoOperator(DefaultProxyTaskToDagsterOperator):
    def default_dagster_run_tags(self, context) -> Mapping[str, Any]:  # pyright: ignore[reportIncompatibleMethodOverride]
        tags = {**super().default_dagster_run_tags(context), MAX_RETRIES_TAG: "3"}
        if self.get_airflow_dag_id(context).endswith("not_step_failure"):
            tags[RETRY_ON_ASSET_OR_OP_FAILURE_TAG] = "false"
        return tags


with DAG(
    dag_id="migrated_asset_has_retries",
    default_args=default_args,
    schedule=None,
    start_date=get_current_datetime_midnight(),
    # We pause this dag upon creation to avoid running it immediately
    is_paused_upon_creation=False,
) as minute_dag:
    PythonOperator(task_id="my_task", python_callable=print_hello)


with DAG(
    dag_id="migrated_asset_has_retries_not_step_failure",
    default_args=default_args,
    schedule=None,
    start_date=get_current_datetime_midnight(),
    # We pause this dag upon creation to avoid running it immediately
    is_paused_upon_creation=False,
) as another_minute_dag:
    PythonOperator(task_id="my_task", python_callable=print_hello)


proxying_to_dagster(
    proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
    global_vars=globals(),
    build_from_task_fn=SetDagsterRetryInfoOperator.build_from_task,
)
